# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/schedulers.fit_flat_varied.ipynb (unless otherwise specified).

__all__ = []

# Cell
import warnings
import numpy as np
from fastcore.basics import even_mults
from fastcore.foundation import patch, L
from fastai.learner import Learner
from fastai.callback.schedule import SchedCos, SchedNo, combine_scheds, ParamScheduler
from ..basics import is_listish

# Cell
@patch
def fit_flat_varied(self:Learner, n_epoch, start_lr=None, div_final=1e5, pct_start=0.75, wd=None,
                    next_lr=None, change_by=None, change_time=1, change_sched=None, cbs=None, reset_opt=False):
    """
    Fit `self.model` for `n_epoch` at flat `start_lr`, then change to flat `next_lr` at `change_by`,
    optionally with cosine annealing or custom `change_sched` over `change_time`. Final cosine annealing at `pct_start`.

    `n_epoch`, `start_lr`, `div_final`, `pct_start`, `wd`, `cbs`, & `reset_opt` are all same as fit_flat_cos from fast.ai.

    `next_lr` single or list of learning rates to switch to at change_by. Must be same length as `change_by`.

    `change_by` single or list of epochs or percent of steps to switch to `next_lr` by. Must be same length as `next_lr`.

    `change_time` if greater than 0 (percent of steps or epochs), how long to cosine anneal to `next_lr`. Can be single or list of same length as `next_lr`.

    `change_sched` optional single or list of fast.ai schedules. If `None` defaults to `SchedCos`. Must be same length as `next_lr`. `SchedPoly` must be passed as partial: `partial(SchedPoly, power=0.5)`.
    """
    assert isinstance(next_lr, (float, slice)) or (is_listish(next_lr) and len(next_lr)>=1), '`next_lr` must be float, slice, or list of float or slice'
    assert isinstance(change_by, (int, float, slice)) or (is_listish(change_by) and len(change_by)>=1), '`change_by` must be int, float, slice, or list of int, float, or slice'

    if self.opt is None: self.create_opt()
    self.opt.set_hyper('lr', self.lr if start_lr is None else start_lr)
    start_lr = np.array([h['lr'] for h in self.opt.hypers])
    params_len = len(start_lr)

    if not is_listish(next_lr): next_lr = [next_lr]
    if not is_listish(change_by): change_by = [change_by]
    change_by = [i/n_epoch if i>=1 else i for i in change_by]
    assert len(change_by)==len(next_lr), '`next_lr` & `change_by` need to be same length'

    if not is_listish(change_time): change_time = [change_time]*len(change_by)
    else: assert len(change_by)==len(change_time), '`change_time` list needs to be same length as `next_lr` & `change_by`'
    change_time = [i/n_epoch if i>=1 else i for i in change_time]

    if change_sched is not None:
        if not is_listish(change_sched): change_sched = [change_sched]
        assert len(change_by)==len(change_sched), '`next_lr` & `change_sched` need to be same length'

    pcts, scheds, last_lr, last_pct = [], [SchedNo(start_lr, start_lr)], start_lr, 0
    for i, cb in enumerate(change_by):
        if cb < pct_start:
            nlr = next_lr[i]
            if isinstance(nlr, slice):
                if nlr.start: nlr = even_mults(nlr.start, nlr.stop, params_len)
                else: nlr = [nlr.stop/10]*(params_len-1) + [nlr.stop]
                nlr=np.array(nlr)
            change_pct = cb - change_time[i]
            assert change_pct >= last_pct, f'{change_pct} in pos {i} of `change_by` overlaps with previous schedule {last_pct}'
            pcts.append(change_pct - sum(pcts))
            scheds.append(SchedNo(nlr, nlr))
            if change_time[i] > 0:
                pcts.append(change_time[i])
                if is_listish(change_sched): scheds.insert(-1, change_sched[i](last_lr, nlr))
                else: scheds.insert(-1, SchedCos(last_lr, nlr))
            last_lr = nlr
            last_pct = change_pct
        else: warnings.warn(f'change_by: {change_by[i]} is after pct_start={pct_start} and ignored.')
    pcts += [pct_start - sum(pcts), 1-pct_start]
    scheds += [SchedCos(last_lr, last_lr/div_final)]

    scheds = {'lr': combine_scheds(pcts, scheds)}
    self.fit(n_epoch, cbs=ParamScheduler(scheds)+L(cbs), reset_opt=reset_opt, wd=wd)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/losses.ipynb (unless otherwise specified).

__all__ = ['ClassBalancedCrossEntropyLoss', 'ClassBalancedBCEWithLogitsLoss']

# Cell
import torch.nn.modules.loss as TL

from .imports import *

# Internal Cell
class ClassBalanced(TL._WeightedLoss):
    "Class Balanced weight calculation, from https://arxiv.org/abs/1901.05555."
    def __init__(self,
        samples_per_class:Tensor,
        beta:float=0.99,
        size_average=None,
        reduce=None,
        reduction:str='mean',
    ):
        num_classes = len(samples_per_class)
        if not isinstance(samples_per_class, Tensor):
            samples_per_class = Tensor(samples_per_class)

        effective_num = 1.0 - torch.pow(beta, samples_per_class)
        weight = (1.0 - beta) / effective_num
        weight = weight / weight.sum() * num_classes

        super().__init__(weight, size_average, reduce, reduction)

    @delegates(nn.Module.to)
    def to(self, *args, **kwargs):
        device, dtype, non_blocking, convert_to_format = torch._C._nn._parse_to(*args, **kwargs)
        if self.weight.device != device: self.weight = self.weight.to(device=device)
        super().to(*args, **kwargs)

# Refactored from https://github.com/vandit15/Class-balanced-loss-pytorch by Vandit Jain

# Original code, Copyright 2019 by Vandit Jain
# Licensed under the MIT License

# Cell
class ClassBalancedCrossEntropyLoss(ClassBalanced):
    "Class Balanced Cross Entropy Loss, from https://arxiv.org/abs/1901.05555."
    __constants__ = ['ignore_index', 'reduction', 'label_smoothing']
    ignore_index: int
    label_smoothing: float

    def __init__(self,
        samples_per_class:Tensor,
        beta:float=0.99,
        size_average=None,
        ignore_index:int=-100,
        reduce=None,
        reduction:str='mean',
        label_smoothing:float=0.0,
        axis:int=-1
    ):
        super().__init__(samples_per_class, beta, size_average, reduce, reduction)
        self.ignore_index = ignore_index
        self.label_smoothing = label_smoothing
        self.axis=axis

    def forward(self, input:Tensor, target:Tensor) -> Tensor:
        return F.cross_entropy(input, target, weight=self.weight,
                               ignore_index=self.ignore_index, reduction=self.reduction,
                               label_smoothing=self.label_smoothing)

    def decodes(self, x:Tensor):    return x.argmax(dim=self.axis)
    def activation(self, x:Tensor): return F.softmax(x, dim=self.axis)

# Cell
class ClassBalancedBCEWithLogitsLoss(ClassBalanced):
    "Class Balanced BCE With Logits Loss, from https://arxiv.org/abs/1901.05555."

    def forward(self, input:Tensor, target:Tensor) -> Tensor:
        return F.binary_cross_entropy_with_logits(input, target, weight=self.weight, reduction=self.reduction)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/test_utils.ipynb (unless otherwise specified).

__all__ = ['test_fail', 'test', 'nequals', 'test_eq', 'test_eq_type', 'test_ne', 'is_close', 'test_close', 'test_is',
           'test_shuffled', 'test_stdout', 'test_warns', 'TEST_IMAGE', 'test_fig_exists', 'ExceptionExpected',
           'exception', 'synth_dbunch', 'RegModel', 'synth_learner', 'VerboseCallback', 'get_env', 'try_import',
           'nvidia_smi', 'nvidia_mem', 'test_show_install', 'less_random', 'TEST_AUDIO']

# Cell
from fastcore.test import (test_fail, test, nequals, test_eq, test_eq_type, test_ne,
                           is_close, test_close, test_is, test_shuffled, test_stdout,
                           test_warns, TEST_IMAGE, test_fig_exists, ExceptionExpected, exception)

from fastai.test_utils import synth_dbunch, synth_learner, RegModel, VerboseCallback, get_env, try_import, nvidia_smi, nvidia_mem
from fastai.test_utils import show_install as test_show_install
from .utils import less_random

# Cell
#nbdev_comment _all_ = ['test_fail', 'test', 'nequals', 'test_eq', 'test_eq_type', 'test_ne', 'is_close', 'test_close', 'test_is', 'test_shuffled', 'test_stdout', 'test_warns', 'TEST_IMAGE', 'test_fig_exists', 'ExceptionExpected', 'exception',  'synth_dbunch', 'RegModel', 'synth_learner', 'VerboseCallback', 'get_env', 'try_import', 'nvidia_smi', 'nvidia_mem', 'test_show_install', 'less_random']

# Cell
TEST_AUDIO = 'audio/whistle.mp3'


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/callback.channelslast.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['ChannelsLastTfm', 'ChannelsLastCallback']

# Cell
#nbdev_comment from __future__ import annotations

from torch.cuda.amp import GradScaler

from fastcore.transform import DisplayedTransform, _is_tuple

from fastai.torch_core import TensorImageBase, TensorMask
from fastai.learner import Learner
from fastai.callback.core import Callback
from fastai.callback.fp16 import MixedPrecision
from fastai.basics import Pipeline

from ..imports import *

# Cell
class ChannelsLastTfm(DisplayedTransform):
    "Sets image-like inputs to `channels_last` format. For use in ChannelsLastCallback"
    order = 110 # run after all other transforms if added to batch_tfms
    def encodes(self, x:TensorImageBase):
        return x.to(memory_format=torch.channels_last)

# Cell
class ChannelsLastCallback(Callback):
    "Channels last training using PyTorch's Channels Last Memory Format (beta)"
    order = 8 # Before CastToTensor
    def __init__(self):
        self._channels_last = Pipeline([ChannelsLastTfm()])

    def before_fit(self):
        self.learn.model.to(memory_format=torch.channels_last)

    def before_batch(self):
        self.learn.xb = self._channels_last(self.xb)

# Cell
@patch
@delegates(GradScaler)
def to_channelslast(self:Learner, to_fp16=True, **kwargs):
    "Set `Learner` and inputs to `channels_last` format and Mixed Precision by default"
    if to_fp16 and not hasattr(self, 'mixed_precision') and not hasattr(self, 'channels_last'):
        return self.add_cbs([ChannelsLastCallback(), MixedPrecision(**kwargs)])
    elif not hasattr(self, 'channels_last'):
        return self.add_cb(ChannelsLastCallback())

# Cell
@patch
def to_contiguous(self:Learner, to_fp32=False):
    "Set `Learner` and inputs to `contiguous_format` (default format), optionally to single precision"
    self.model.to(memory_format=torch.contiguous_format)
    if to_fp32: return self.remove_cbs([ChannelsLastCallback, MixedPrecision])
    else:       return self.remove_cb(ChannelsLastCallback)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/data.transforms.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['KFoldColSplitter', 'ParentSplitter', 'GreatGrandparentSplitter']

# Cell
#nbdev_comment from __future__ import annotations

import pandas as pd
from pathlib import Path

from fastcore.foundation import mask2idxs

from fastai.data.transforms import IndexSplitter

from ..imports import *

# Cell
def KFoldColSplitter(
    fold:listified[int]=0, # Valid set fold(s)
    col:int|str='folds' # Column with folds
):
    "Split `items` (supposed to be a dataframe) by `fold` in `col`"
    def _inner(o):
        assert isinstance(o, pd.DataFrame), "KFoldColSplitter only works when your items are a pandas DataFrame"
        valid_col = o.iloc[:,col] if isinstance(col, int) else o[col]
        valid_idx = valid_col.isin(fold) if is_listy(fold) else valid_col.values == fold
        return IndexSplitter(mask2idxs(valid_idx))(o)
    return _inner

# Internal Cell
def _parent_idxs(items, name):
    def _inner(items, name): return mask2idxs(Path(o).parent.name == name for o in items)
    return [i for n in L(name) for i in _inner(items,n)]

# Cell
def ParentSplitter(
    train_name:str='train', # Train set folder name
    valid_name:str='valid' # Valid set folder name
):
    "Split `items` from the parent folder names (`train_name` and `valid_name`)."
    def _inner(o):
        return _parent_idxs(o, train_name),_parent_idxs(o, valid_name)
    return _inner

# Internal Cell
def _greatgrandparent_idxs(items, name):
    def _inner(items, name): return mask2idxs(Path(o).parent.parent.parent.name == name for o in items)
    return [i for n in L(name) for i in _inner(items,n)]

# Cell
def GreatGrandparentSplitter(
    train_name:str='train', # Train set folder name
    valid_name:str='valid' # Valid set folder name
):
    "Split `items` from the great grand parent folder names (`train_name` and `valid_name`)."
    def _inner(o):
        return _greatgrandparent_idxs(o, train_name),_greatgrandparent_idxs(o, valid_name)
    return _inner


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/multiloss.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['MultiLoss', 'MultiTargetLoss', 'MultiLossCallback']

# Cell
#nbdev_comment from __future__ import annotations
from types import FunctionType

from fastai.callback.core import Callback
from fastai.learner import Recorder

from .basics import is_listish
from .metrics import AvgLossX, AvgSmoothLossX
from .imports import *

# Internal Cell
def init_loss(l, **kwargs):
    "Initiatize loss class or partial loss function"
    return partialler(l, reduction='none') if isinstance(l, FunctionType) else l(reduction='none', **kwargs)

# Cell
class MultiLoss(Module):
    """
    Combine multiple `loss_funcs` on one prediction & target via `reduction`, with optional weighting.

    Log `loss_funcs` as metrics via `MultiLossCallback`, optionally using `loss_names`.
    """
    def __init__(self,
        loss_funcs:listy[Callable[...,nn.Module]|FunctionType], # Uninitialized loss functions or classes. Must support PyTorch `reduction` string.
        weights:listified[Number]|None=None, # Weight per loss. Defaults to uniform weighting.
        loss_kwargs:listy[dict[str,Any]]|None=None, # kwargs to pass to each loss function. Defaults to None.
        loss_names:listy[str]|None=None, # Loss names to log using `MultiLossCallback`. Defaults to loss `__name__`.
        reduction:str|None='mean' # PyTorch loss reduction
    ):
        store_attr(but='loss_names')
        assert is_listy(loss_funcs), "`loss_funcs` must be list-like"

        if weights is None or len(weights)==0:
            self.weights = [1]*len(loss_funcs)
        else:
            assert len(loss_funcs) == len(weights), "Must provide same number of `weights` as `loss_funcs`"
            self.weights = weights

        if loss_kwargs is None or len(loss_kwargs)==0: loss_kwargs = [{}]*len(loss_funcs)
        else: assert len(loss_funcs) == len(loss_kwargs), "Must provide same number of `loss_kwargs` as `loss_funcs`"

        if loss_names is None or len(loss_names)==0: loss_names = [l.__name__ for l in loss_funcs]
        else: assert len(loss_funcs) == len(loss_names), "Must provide same number of `loss_names` as `loss_funcs`"

        self.loss_funcs = [init_loss(l, **k) for l, k in zip(loss_funcs, loss_kwargs)]
        self.loss_names = L(loss_names)
        self._reduction,self._loss = reduction,{}

    def forward(self, pred, targ):
        for i, loss_func in enumerate(self.loss_funcs):
            l = TensorBase(self.weights[i]*loss_func(pred, targ))
            if i == 0: loss = TensorBase(torch.zeros_like(targ)).float()
            loss += l
            self._loss[i] = l

        return loss.mean() if self._reduction=='mean' else loss.sum() if self._reduction=='sum' else loss

    @property
    def losses(self): return self._loss

    @property
    def reduction(self): return self._reduction

    @reduction.setter
    def reduction(self, r): self._reduction = r

    @delegates(Module.to)
    def to(self, *args, **kwargs):
        device, dtype, non_blocking, convert_to_format = torch._C._nn._parse_to(*args, **kwargs)
        if is_listish(self.weights) or not isinstance(self.weights, torch.Tensor): self.weights = torch.Tensor(self.weights)
        if self.weights.device != device: self.weights = self.weights.to(device=device)
        super().to(*args, **kwargs)

    def activation(self, pred):
        "Returns first `loss_funcs` `activation`"
        return getattr(self.loss_funcs[0], 'activation', noop)(pred)

    def decodes(self, pred):
        "Returns first `loss_funcs` `decodes`"
        return getattr(self.loss_funcs[0], 'decodes', noop)(pred)

# Cell
class MultiTargetLoss(MultiLoss):
    """
    Combine `loss_funcs` from multiple predictions & targets via `reduction`, with optional weighting.

    Log `loss_funcs` as metrics via `MultiLossCallback`, optionally using `loss_names`.
    """
    def __init__(self,
        loss_funcs:listy[Callable[...,nn.Module]|FunctionType], # Uninitialized loss functions or classes. One per prediction and target. Must support PyTorch `reduction` string.
        weights:listified[Number]|None=None, # Weight per loss. Defaults to uniform weighting.
        loss_kwargs:listy[dict[str,Any]]|None=None, # kwargs to pass to each loss function. Defaults to None.
        loss_names:listy[str]|None=None, # Loss names to log using `MultiLossCallback`. Defaults to loss `__name__`.
        reduction:str|None='mean' # PyTorch loss reduction
    ):
        super().__init__(loss_funcs, weights, loss_kwargs, loss_names, reduction)

    def forward(self, preds, targs):
        for i, (loss_func, pred, targ) in enumerate(zip(self.loss_funcs, preds, targs)):
            l = TensorBase(self.weights[i]*loss_func(pred, targ))
            if i == 0: loss = TensorBase(torch.zeros_like(targ)).float()
            loss += l
            self._loss[i] = l

        return loss.mean() if self._reduction=='mean' else loss.sum() if self._reduction=='sum' else loss

    def activation(self, preds):
        "Returns list of `activation`"
        return [getattr(self.loss_funcs[i], 'activation', noop)(pred) for i, pred in enumerate(preds)]

    def decodes(self, preds):
        "Returns list of `decodes`"
        return [getattr(self.loss_funcs[i], 'decodes', noop)(pred) for i, pred in enumerate(preds)]

# Internal Cell
class MultiAvgLoss(AvgLossX):
    "Average the MultiLoss losses taking into account potential different batch sizes"
    def __init__(self,
        i, # `Multiloss` loss function location
        name, # Loss function name
        reduction:str|None='mean' # Override loss reduction for logging
    ):
        store_attr(but='name')
        self._name = name

    def accumulate(self, learn):
        bs = find_bs(learn.yb)
        loss = learn.loss_func.losses[self.i]
        loss = loss.mean() if self.reduction=='mean' else loss.sum() if self.reduction=='sum' else loss
        self.total += learn.to_detach(loss)*bs
        self.count += bs

# Internal Cell
class MultiAvgSmoothLoss(AvgSmoothLossX):
    "Smooth average of the MultiLoss losses (exponentially weighted with `beta`)"
    def __init__(self,
        i, # `Multiloss` loss function location
        name, # Loss function name
        beta:float=0.98, # Smoothing beta
        reduction:str|None='mean' # Override loss reduction for logging
    ):
        super().__init__()
        store_attr(but='name')
        self._name = name

    def accumulate(self, learn):
        self.count += 1
        loss = learn.loss_func.losses[self.i]
        loss = loss.mean() if self.reduction=='mean' else loss.sum() if self.reduction=='sum' else loss
        self.val = torch.lerp(to_detach(loss, gather=False), self.val, self.beta)

# Cell
class MultiLossCallback(Callback):
    "Callback to automatically log and name `MultiLoss` losses as fastxtend metrics"
    run_valid,order = False,Recorder.order-1

    def __init__(self,
        beta:float=0.98, # Smoothing beta
        reduction:str|None='mean' # Override loss reduction for logging
    ):
        store_attr()

    def before_fit(self):
        if not isinstance(self.loss_func, MultiLoss):
            raise ValueError("`MultiLossCallback` requires loss to be `MultiLoss` class")

        mets= L()
        reduction = self.loss_func.reduction if self.reduction is None else self.reduction
        for i in range(len(self.loss_func.loss_funcs)):
            mets += MultiAvgSmoothLoss(i, self.loss_func.loss_names[i], self.beta, reduction)
            mets += MultiAvgLoss(i, self.loss_func.loss_names[i], reduction)

        self.learn.metrics = mets + self.learn.metrics


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/metrics.ipynb (unless otherwise specified).

__all__ = ['LogMetric', 'MetricType', 'ActivationType', 'MetricX', 'AvgMetricX', 'AccumMetricX', 'AvgSmoothMetricX',
           'AvgLossX', 'AvgSmoothLossX', 'ValueMetricX', 'func_to_metric', 'skm_to_fastxtend', 'Accuracy', 'ErrorRate',
           'TopKAccuracy', 'APScoreBinary', 'BalancedAccuracy', 'BrierScore', 'CohenKappa', 'F1Score', 'FBeta',
           'HammingLoss', 'Jaccard', 'Precision', 'Recall', 'RocAuc', 'RocAucBinary', 'MatthewsCorrCoef',
           'AccuracyMulti', 'APScoreMulti', 'BrierScoreMulti', 'F1ScoreMulti', 'FBetaMulti', 'HammingLossMulti',
           'JaccardMulti', 'MatthewsCorrCoefMulti', 'PrecisionMulti', 'RecallMulti', 'RocAucMulti', 'MSE', 'RMSE',
           'MAE', 'MSLE', 'ExpRMSE', 'ExplainedVariance', 'R2Score', 'PearsonCorrCoef', 'SpearmanCorrCoef',
           'ForegroundAcc', 'Dice', 'DiceMulti', 'JaccardCoeff', 'CorpusBLEUMetric', 'Perplexity', 'perplexity',
           'LossMetric', 'LossMetrics']

# Cell
import inspect, time, math

import numpy as np
import sklearn.metrics as skm
import scipy.stats as scs

from fastcore.basics import class2attr
from fastcore.dispatch import cast

from fastprogress.core import format_time

from fastai.callback.core import Callback
from fastai.learner import AvgMetric, Learner, Metric, Recorder, _maybe_item
from fastai.metrics import AccumMetric
from fastai.torch_core import flatten_check

from .imports import *

# Cell
class LogMetric(Enum):
    "All possible logging types for `MetricX`",
    Train = 1
    Valid = 2
    Both = 3

# Cell
class MetricType(Enum):
    "All possible types of `MetricX`",
    Avg = 1
    Accum = 2
    Smooth = 3

# Cell
class ActivationType(Enum):
    "All possible activation classes for `MetricX",
    No = 1
    Sigmoid = 2
    Softmax = 3
    BinarySoftmax = 4

# Cell
class MetricX(Metric):
    "Blueprint for defining an extended metric with accumulate"
    log_metric=LogMetric.Valid
    def __init__(self, dim_argmax=None, activation=ActivationType.No, thresh=None, log_metric=None, name=None):
        store_attr(but='log_metric, name')
        self.log_metric = ifnone(log_metric, self.log_metric)
        self._name = ifnone(name, class2attr(self, 'MetricX'))

    def reset(self):
        "Reset inner state to prepare for new computation"
        pass

    def accumulate(self, learn):
        "Store targs and preds from `learn`, using activation function and argmax as appropriate"
        self.pred, self.targ = learn.pred, *learn.yb
        if self.activation in [ActivationType.Softmax, ActivationType.BinarySoftmax]:
            self.pred = F.softmax(self.pred, dim=self.dim_argmax)
            if self.activation == ActivationType.BinarySoftmax: self.pred = self.pred[:, -1]
        elif self.activation == ActivationType.Sigmoid: self.pred = torch.sigmoid(self.pred)
        elif self.dim_argmax: self.pred = self.pred.argmax(dim=self.dim_argmax)
        if self.thresh: self.pred = (self.pred >= self.thresh)

    @property
    def value(self):
        "The value of the metric"
        raise NotImplementedError

    @property
    def name(self):
        "Name of the `Metric`, camel-cased and with Metric removed. Or custom name if provided"
        return self._name

    @name.setter
    def name(self, value): self._name = value

    def _split_kwargs(self, method, **kwargs):
        args = [k for k,v in inspect.signature(method).parameters.items() if v.default != inspect.Parameter.empty]
        return {k: kwargs[k] for k in kwargs.keys() if k in args}

# Cell
@delegates(MetricX)
class AvgMetricX(MetricX):
    "Average the values of `func` taking into account potential different batch sizes"
    def __init__(self, func, to_np=False, invert_arg=False, **kwargs):
        super().__init__(**self._split_kwargs(MetricX.__init__, **kwargs))
        self.func, self.fkwargs = func, self._split_kwargs(func, **kwargs)
        self.to_np, self.invert_arg = to_np, invert_arg
        self._name = ifnone(kwargs.get('name', None), self.func.func.__name__ if hasattr(self.func, 'func') else self.func.__name__)

    def reset(self): self.total,self.count = 0.,0

    def accumulate(self, learn):
        super().accumulate(learn)
        bs = find_bs(learn.yb)
        if self.to_np: self.pred,self.targ = learn.to_detach(self.pred).numpy(),learn.to_detach(self.targ).numpy()
        self.total += (self.func(self.targ, self.pred, **self.fkwargs) if self.invert_arg else self.func(self.pred, self.targ, **self.fkwargs))*bs
        self.count += bs

    @property
    def value(self): return self.total/self.count if self.count != 0 else None

# Cell
@delegates(MetricX)
class AccumMetricX(MetricX):
    "Stores predictions and targets on CPU in accumulate to perform final calculations with `func`."
    def __init__(self, func, to_np=False, invert_arg=False, flatten=True, **kwargs):
        super().__init__(**self._split_kwargs(MetricX.__init__, **kwargs))
        self.flatten, self.func, self.fkwargs = flatten, func, self._split_kwargs(func, **kwargs)
        self.to_np, self.invert_arg = to_np, invert_arg
        self._name = ifnone(kwargs.get('name', None), self.func.func.__name__ if hasattr(self.func, 'func') else self.func.__name__)

    def reset(self):
        "Clear all targs and preds"
        self.targs,self.preds = [],[]

    def accumulate(self, learn):
        "Store targs and preds from `learn`, using activation function and argmax as appropriate"
        super().accumulate(learn)
        self.pred,self.targ = learn.to_detach(self.pred),learn.to_detach(self.targ)
        self.accum_values(self.pred, self.targ)

    def accum_values(self, preds, targs):
        "Store targs and preds"
        if self.flatten: preds,targs = flatten_check(preds,targs)
        self.preds.append(preds)
        self.targs.append(targs)

    def __call__(self, preds, targs):
        "Calculate metric on one batch of data"
        self.reset()
        self.accum_values(preds,targs)
        return self.value

    @property
    def value(self):
        "Value of the metric using accumulated preds and targs"
        if len(self.preds) == 0: return
        preds,targs = torch.cat(self.preds),torch.cat(self.targs)
        if self.to_np: preds,targs = preds.numpy(),targs.numpy()
        return self.func(targs, preds, **self.fkwargs) if self.invert_arg else self.func(preds, targs, **self.fkwargs)

# Cell
@delegates(MetricX, but='log_metric')
class AvgSmoothMetricX(MetricX):
    "Smooth average the values of `func` (exponentially weighted with `beta`). Only computed on training set."
    log_metric = LogMetric.Train
    def __init__(self, func, beta=0.98, to_np=False, invert_arg=False, **kwargs):
        super().__init__(**self._split_kwargs(MetricX.__init__, **kwargs))
        self.func, self.fkwargs = func, self._split_kwargs(func, **kwargs)
        self.beta, self.to_np, self.invert_arg = beta, to_np, invert_arg
        self._name = ifnone(kwargs.get('name', None), self.func.func.__name__ if hasattr(self.func, 'func') else self.func.__name__)

    def reset(self): self.count,self.val = 0,tensor(0.)

    def accumulate(self, learn):
        super().accumulate(learn)
        if self.to_np: self.pred,self.targ = learn.to_detach(self.pred).numpy(),learn.to_detach(self.targ).numpy()
        val = self.func(self.targ, self.pred, **self.fkwargs) if self.invert_arg else self.func(self.pred, self.targ, **self.fkwargs)
        if self.to_np: self.val = val*self.beta + val*(1-self.beta)
        else: self.val = torch.lerp(to_detach(val, gather=False), self.val, self.beta)
        self.count += 1

    @property
    def value(self): return self.val/(1-self.beta**self.count) if self.count != 0 else None

# Cell
class AvgLossX(MetricX):
    "Average the losses taking into account potential different batch sizes"
    def reset(self): self.total,self.count = 0.,0
    def accumulate(self, learn):
        bs = find_bs(learn.yb)
        self.total += learn.to_detach(learn.loss.mean())*bs
        self.count += bs
    @property
    def value(self): return self.total/self.count if self.count != 0 else None

# Cell
class AvgSmoothLossX(MetricX):
    "Smooth average of the losses (exponentially weighted with `beta`)"
    log_metric = LogMetric.Train
    def __init__(self, beta=0.98):
        self.beta = beta
    def reset(self):
        self.count,self.val = 0,tensor(0.)
    def accumulate(self, learn):
        self.count += 1
        self.val = torch.lerp(to_detach(learn.loss.mean(), gather=False), self.val, self.beta)
    @property
    def value(self): return self.val/(1-self.beta**self.count)

# Cell
class ValueMetricX(MetricX):
    "Use to include a pre-calculated metric value (for instance calculated in a `Callback`) and returned by `func`"
    def __init__(self, func, name=None, log_metric=None):
        super().__init__(log_metric=log_metric)
        self.func = func
        self._name = ifnone(name, self.func.func.__name__ if hasattr(self.func, 'func') else self.func.__name__)

    @property
    def value(self): return self.func()

# Internal Cell
def _dedup_metric_names(metrics, names):
    dup = set()
    log = metrics.map(lambda o: (LogMetric.Valid if o.log_metric==LogMetric.Both else o.log_metric) if hasattr(o, 'log_metric') else LogMetric.Valid)
    dups = L(set([o[1] for o in zip(log, names) if o in dup or dup.add(o)]))
    indices = names.argwhere(lambda o: o in dups)
    for i in indices:
        if hasattr(metrics[i], 'func'):
            name = metrics[i].func.func.__name__ if hasattr(metrics[i].func, 'func') else metrics[i].func.__name__
        else:
            if isinstance(metrics[i], MetricX):
                name = class2attr(metrics[i], 'MetricX')
            else:
                name = class2attr(metrics[i], 'Metric')
        if not hasattr(metrics[i], 'name') or metrics[i].name == name: # only deduplicate default metric names
            if isinstance(metrics[i], (AvgMetricX, AvgMetric)):
                names[i] = f'avg_{names[i]}'
            elif isinstance(metrics[i], (AccumMetricX, AccumMetric)):
                names[i] = f'accm_{names[i]}'
            elif isinstance(metrics[i], AvgSmoothMetricX):
                names[i] = f'smth_{names[i]}'
    return names

# Internal Cell
@patch
def __init__(self:Recorder, add_time=True, train_metrics=False, valid_metrics=True, beta=0.98):
    store_attr('add_time')
    self.loss,self.smooth_loss = AvgLossX(),AvgSmoothLossX(beta=beta)

@patch
def before_fit(self:Recorder):
    "Prepare state for training"
    self.lrs,self.iters,self.losses,self.values = [],[],[],[]
    names = self.metrics.attrgot('name')
    if len(names.unique()) != len(names):
        names = _dedup_metric_names(self.metrics, names)
    train = self.metrics.argwhere(lambda o: hasattr(o, 'log_metric') and o.log_metric != LogMetric.Valid)
    valid = self.metrics.argwhere(lambda o: not hasattr(o, 'log_metric') or o.log_metric != LogMetric.Train)
    self._train_metsX = self.metrics[train]
    self._valid_metsX = self.metrics[valid]
    train_names = names[train]
    valid_names = names[valid]
    if len(self._train_metsX) > 0:
        train_names = train_names.map('train_{}')
        valid_names = valid_names.map('valid_{}')
    smooth = self._train_metsX.argwhere(lambda o: isinstance(o, (AvgSmoothLossX, AvgSmoothMetricX)))
    self.smooth_mets  = self._train_metsX[smooth]
    self.smooth_names = train_names[smooth]
    self.train_names = L('train_loss') + train_names
    self.valid_names = L('valid_loss') + valid_names
    self.metric_names = L('epoch') + self.train_names + self.valid_names
    if self.add_time: self.metric_names.append('time')
    self.smooth_loss.reset()
    self.loss.reset()
    self.smooth_mets.map(Self.reset())

@patch
def after_batch(self:Recorder):
    "Update all metrics and records lr and smooth loss in training"
    if len(self.yb) == 0: return
    mets = self.train_mets() if self.training else self.valid_mets()
    for met in mets:
        met.accumulate(self.learn)
    if not self.training: return
    self.lrs.append(self.opt.hypers[-1]['lr'])
    self.losses.append(self.smooth_loss.value)
    self.learn.smooth_loss = self.smooth_loss.value

@patch
def before_epoch(self:Recorder):
    "Set timer if `self.add_time=True`"
    self.cancel_train,self.cancel_valid = False,False
    if self.add_time: self.start_epoch = time.time()
    self.log = L(getattr(self, 'epoch', 0))

@patch
def before_train(self:Recorder):
    self._train_metsX.filter(lambda o: not isinstance(o, (AvgSmoothLossX, AvgSmoothMetricX))).map(Self.reset())

@patch
def before_validate(self:Recorder):
    self.valid_mets().map(Self.reset())

@patch
def after_train(self:Recorder):
    self.log += self.train_mets().map(_maybe_item)

@patch
def after_validate(self:Recorder):
    self.log += self.valid_mets().map(_maybe_item)

@patch
def after_epoch(self:Recorder):
    "Store and log the loss/metric values"
    self.learn.final_record = self.log[1:].copy()
    self.values.append(self.learn.final_record)
    if self.add_time: self.log.append(format_time(time.time() - self.start_epoch))
    self.logger(self.log)
    self.iters.append(self.smooth_loss.count)

@patch
def train_mets(self:Recorder):
    if getattr(self, 'cancel_train', False): return L()
    return L(self.smooth_loss) + self._train_metsX

@patch
def valid_mets(self:Recorder):
    if getattr(self, 'cancel_valid', False): return L()
    return L(self.loss) + self._valid_metsX

# Cell
@delegates(MetricX)
def func_to_metric(func, metric_type, is_class, thresh=None, axis=-1, activation=None, log_metric=LogMetric.Valid, **kwargs):
    "Convert `func` metric to a fastai metric"

    dim_argmax = axis if is_class and thresh is None else None
    if activation is None:
        activation = ActivationType.Sigmoid if (is_class and thresh is not None) else ActivationType.No

    if metric_type==MetricType.Accum:
        return AccumMetricX(func, dim_argmax=dim_argmax, activation=activation,
                            thresh=thresh, log_metric=log_metric, **kwargs)
    elif metric_type==MetricType.Avg:
        return AvgMetricX(func, dim_argmax=dim_argmax, activation=activation,
                          thresh=thresh, log_metric=log_metric, **kwargs)
    elif metric_type==MetricType.Smooth:
        if log_metric!=LogMetric.Train:
            name = func.func.__name__ if hasattr(func, 'func') else  func.__name__
            raise ValueError(f'Error with {name}: AvgSmoothMetricX can only run on train. Set `log_metric` to LogMetric.Train.')
        return AvgSmoothMetricX(func, dim_argmax=dim_argmax, activation=activation, thresh=thresh, **kwargs)
    else:
        name = func.func.__name__ if hasattr(func, 'func') else  func.__name__
        raise ValueError(f"Unsupported `metric_type` {metric_type} for metric {name}.")

# Cell
@delegates(MetricX)
def skm_to_fastxtend(func, is_class=True, thresh=None, axis=-1, activation=None, log_metric=LogMetric.Valid, **kwargs):
    "Convert `func` from sklearn.metrics to a fastai metric"
    return func_to_metric(func, MetricType.Accum, is_class, thresh, axis, activation,
                          log_metric, to_np=True, invert_arg=True, **kwargs)

# Internal Cell
def accuracy(inp, targ):
    "Compute accuracy with `targ` when `pred` is bs * n_classes"
    pred,targ = flatten_check(inp, targ)
    return (pred == targ).float().mean()

# Cell
def Accuracy(axis=-1, metric_type=MetricType.Avg, log_metric=LogMetric.Valid, **kwargs):
    "Compute accuracy with `targ` when `pred` is bs * n_classes"
    return func_to_metric(accuracy, metric_type, True, axis=axis, log_metric=log_metric, **kwargs)

# Internal Cell
def error_rate(inp, targ):
    "1 - `accuracy`"
    return 1 - accuracy(inp, targ)

# Cell
def ErrorRate(axis=-1, metric_type=MetricType.Avg, log_metric=LogMetric.Valid, **kwargs):
    "Compute 1 - accuracy with `targ` when `pred` is bs * n_classes"
    return func_to_metric(error_rate, metric_type, True, axis=axis, log_metric=log_metric, **kwargs)

# Internal Cell
def top_k_accuracy(inp, targ, k=5, axis=-1):
    "Computes the Top-k accuracy (`targ` is in the top `k` predictions of `inp`)"
    inp = inp.topk(k=k, dim=axis)[1]
    targ = targ.unsqueeze(dim=axis).expand_as(inp)
    return (inp == targ).sum(dim=-1).float().mean()

# Cell
def TopKAccuracy(k=5, axis=-1, metric_type=MetricType.Avg, log_metric=LogMetric.Valid, **kwargs):
    "Computes the Top-k accuracy (`targ` is in the top `k` predictions of `inp`)"
    return func_to_metric(partial(top_k_accuracy, k=k, axis=axis), metric_type, False,
                          log_metric=log_metric, **kwargs)

# Cell
def APScoreBinary(axis=-1, average='macro', pos_label=1, sample_weight=None, log_metric=LogMetric.Valid, **kwargs):
    "Average Precision for single-label binary classification problems"
    return skm_to_fastxtend(skm.average_precision_score, axis=axis, activation=ActivationType.BinarySoftmax,
                         average=average, pos_label=pos_label, sample_weight=sample_weight, log_metric=log_metric,
                         **kwargs)

# Cell
def BalancedAccuracy(axis=-1, sample_weight=None, adjusted=False, log_metric=LogMetric.Valid, **kwargs):
    "Balanced Accuracy for single-label binary classification problems"
    return skm_to_fastxtend(skm.balanced_accuracy_score, axis=axis,
                         sample_weight=sample_weight, adjusted=adjusted, log_metric=log_metric, **kwargs)

# Cell
def BrierScore(axis=-1, sample_weight=None, pos_label=None, log_metric=LogMetric.Valid, **kwargs):
    "Brier score for single-label classification problems"
    return skm_to_fastxtend(skm.brier_score_loss, axis=axis,
                         sample_weight=sample_weight, pos_label=pos_label, log_metric=log_metric, **kwargs)

# Cell
def CohenKappa(axis=-1, labels=None, weights=None, sample_weight=None, log_metric=LogMetric.Valid, **kwargs):
    "Cohen kappa for single-label classification problems"
    return skm_to_fastxtend(skm.cohen_kappa_score, axis=axis, labels=labels, weights=weights,
                         sample_weight=sample_weight, log_metric=log_metric, **kwargs)

# Cell
def F1Score(axis=-1, labels=None, pos_label=1, average='binary', sample_weight=None, log_metric=LogMetric.Valid, **kwargs):
    "F1 score for single-label classification problems"
    return skm_to_fastxtend(skm.f1_score, axis=axis, labels=labels, pos_label=pos_label,
                         average=average, sample_weight=sample_weight, log_metric=log_metric, **kwargs)

# Cell
def FBeta(beta, axis=-1, labels=None, pos_label=1, average='binary', sample_weight=None,
          log_metric=LogMetric.Valid, **kwargs):
    "FBeta score with `beta` for single-label classification problems"
    return skm_to_fastxtend(skm.fbeta_score, axis=axis, beta=beta, labels=labels, pos_label=pos_label,
                         average=average, sample_weight=sample_weight, log_metric=log_metric, **kwargs)

# Cell
def HammingLoss(axis=-1, sample_weight=None, log_metric=LogMetric.Valid, **kwargs):
    "Hamming loss for single-label classification problems"
    return skm_to_fastxtend(skm.hamming_loss, axis=axis,
                         sample_weight=sample_weight, log_metric=log_metric, **kwargs)

# Cell
def Jaccard(axis=-1, labels=None, pos_label=1, average='binary', sample_weight=None,
            log_metric=LogMetric.Valid, **kwargs):
    "Jaccard score for single-label classification problems"
    return skm_to_fastxtend(skm.jaccard_score, axis=axis, labels=labels, pos_label=pos_label,
                         average=average, sample_weight=sample_weight, log_metric=log_metric, **kwargs)

# Cell
def Precision(axis=-1, labels=None, pos_label=1, average='binary', sample_weight=None,
              log_metric=LogMetric.Valid, **kwargs):
    "Precision for single-label classification problems"
    return skm_to_fastxtend(skm.precision_score, axis=axis, labels=labels, pos_label=pos_label,
                         average=average, sample_weight=sample_weight, log_metric=log_metric, **kwargs)

# Cell
def Recall(axis=-1, labels=None, pos_label=1, average='binary', sample_weight=None,
           log_metric=LogMetric.Valid, **kwargs):
    "Recall for single-label classification problems"
    return skm_to_fastxtend(skm.recall_score, axis=axis, labels=labels, pos_label=pos_label,
                         average=average, sample_weight=sample_weight, log_metric=log_metric, **kwargs)

# Cell
def RocAuc(axis=-1, average='macro', sample_weight=None, max_fpr=None, multi_class='ovr',
           log_metric=LogMetric.Valid, **kwargs):
    "Area Under the Receiver Operating Characteristic Curve for single-label multiclass classification problems"
    assert multi_class in ['ovr', 'ovo']
    return skm_to_fastxtend(skm.roc_auc_score, axis=axis, activation=ActivationType.Softmax, flatten=False,
                         average=average, sample_weight=sample_weight, max_fpr=max_fpr, multi_class=multi_class,
                         log_metric=log_metric, **kwargs)

# Cell
def RocAucBinary(axis=-1, average='macro', sample_weight=None, max_fpr=None, multi_class='raise',
                 log_metric=LogMetric.Valid, **kwargs):
    "Area Under the Receiver Operating Characteristic Curve for single-label binary classification problems"
    return skm_to_fastxtend(skm.roc_auc_score, axis=axis, activation=ActivationType.BinarySoftmax,
                         average=average, sample_weight=sample_weight, max_fpr=max_fpr, multi_class=multi_class,
                         log_metric=log_metric, **kwargs)

# Cell
def MatthewsCorrCoef(sample_weight=None, log_metric=LogMetric.Valid, **kwargs):
    "Matthews correlation coefficient for single-label classification problems"
    return skm_to_fastxtend(skm.matthews_corrcoef, sample_weight=sample_weight, log_metric=log_metric, **kwargs)

# Internal Cell
def accuracy_multi(inp, targ):
    "Compute accuracy when `inp` and `targ` are the same size."
    inp,targ = flatten_check(inp,targ)
    return (inp==targ.bool()).float().mean()

# Cell
def AccuracyMulti(thresh=0.5, sigmoid=True, metric_type=MetricType.Avg, log_metric=LogMetric.Valid, **kwargs):
    "Compute accuracy when `inp` and `targ` are the same size."
    activation = ActivationType.Sigmoid if sigmoid else ActivationType.No
    return func_to_metric(accuracy_multi, metric_type, False, thresh=thresh, activation=activation,
                          log_metric=log_metric, **kwargs)

# Cell
def APScoreMulti(sigmoid=True, average='macro', pos_label=1, sample_weight=None,
                 log_metric=LogMetric.Valid, **kwargs):
    "Average Precision for multi-label classification problems"
    activation = ActivationType.Sigmoid if sigmoid else ActivationType.No
    return skm_to_fastxtend(skm.average_precision_score, activation=activation, flatten=False,
                         average=average, pos_label=pos_label, sample_weight=sample_weight,
                         log_metric=log_metric, **kwargs)

# Cell
def BrierScoreMulti(thresh=0.5, sigmoid=True, sample_weight=None, pos_label=None,
                    log_metric=LogMetric.Valid, **kwargs):
    "Brier score for multi-label classification problems"
    activation = ActivationType.Sigmoid if sigmoid else ActivationType.No
    return skm_to_fastxtend(skm.brier_score_loss, thresh=thresh, activation=activation, flatten=False,
                         sample_weight=sample_weight, pos_label=pos_label, log_metric=log_metric, **kwargs)

# Cell
def F1ScoreMulti(thresh=0.5, sigmoid=True, labels=None, pos_label=1, average='macro', sample_weight=None,
                 log_metric=LogMetric.Valid, **kwargs):
    "F1 score for multi-label classification problems"
    activation = ActivationType.Sigmoid if sigmoid else ActivationType.No
    return skm_to_fastxtend(skm.f1_score, thresh=thresh, activation=activation, flatten=False,
                         labels=labels, pos_label=pos_label, average=average, sample_weight=sample_weight,
                         log_metric=log_metric, **kwargs)

# Cell
def FBetaMulti(beta, thresh=0.5, sigmoid=True, labels=None, pos_label=1, average='macro', sample_weight=None,
               log_metric=LogMetric.Valid, **kwargs):
    "FBeta score with `beta` for multi-label classification problems"
    activation = ActivationType.Sigmoid if sigmoid else ActivationType.No
    return skm_to_fastxtend(skm.fbeta_score, thresh=thresh, activation=activation, flatten=False,
                beta=beta, labels=labels, pos_label=pos_label, average=average, sample_weight=sample_weight,
                log_metric=log_metric, **kwargs)

# Cell
def HammingLossMulti(thresh=0.5, sigmoid=True, labels=None, sample_weight=None,
                     log_metric=LogMetric.Valid, **kwargs):
    "Hamming loss for multi-label classification problems"
    activation = ActivationType.Sigmoid if sigmoid else ActivationType.No
    return skm_to_fastxtend(skm.hamming_loss, thresh=thresh, activation=activation, flatten=False,
                         sample_weight=sample_weight, log_metric=log_metric, **kwargs)

# Cell
def JaccardMulti(thresh=0.5, sigmoid=True, labels=None, pos_label=1, average='macro', sample_weight=None,
                 log_metric=LogMetric.Valid, **kwargs):
    "Jaccard score for multi-label classification problems"
    activation = ActivationType.Sigmoid if sigmoid else ActivationType.No
    return skm_to_fastxtend(skm.jaccard_score, thresh=thresh, activation=activation, flatten=False,
                         labels=labels, pos_label=pos_label, average=average, sample_weight=sample_weight,
                         log_metric=log_metric, **kwargs)

# Cell
def MatthewsCorrCoefMulti(thresh=0.5, sigmoid=True, sample_weight=None, log_metric=LogMetric.Valid, **kwargs):
    "Matthews correlation coefficient for multi-label classification problems"
    activation = ActivationType.Sigmoid if sigmoid else ActivationType.No
    return skm_to_fastxtend(skm.matthews_corrcoef, thresh=thresh, activation=activation, flatten=False,
                         sample_weight=sample_weight, log_metric=log_metric, **kwargs)

# Cell
def PrecisionMulti(thresh=0.5, sigmoid=True, labels=None, pos_label=1, average='macro', sample_weight=None,
                   log_metric=LogMetric.Valid, **kwargs):
    "Precision for multi-label classification problems"
    activation = ActivationType.Sigmoid if sigmoid else ActivationType.No
    return skm_to_fastxtend(skm.precision_score, thresh=thresh, activation=activation, flatten=False,
                         labels=labels, pos_label=pos_label, average=average, sample_weight=sample_weight,
                         log_metric=log_metric, **kwargs)

# Cell
def RecallMulti(thresh=0.5, sigmoid=True, labels=None, pos_label=1, average='macro', sample_weight=None,
                log_metric=LogMetric.Valid, **kwargs):
    "Recall for multi-label classification problems"
    activation = ActivationType.Sigmoid if sigmoid else ActivationType.No
    return skm_to_fastxtend(skm.recall_score, thresh=thresh, activation=activation, flatten=False,
                         labels=labels, pos_label=pos_label, average=average, sample_weight=sample_weight,
                         log_metric=log_metric, **kwargs)

# Cell
def RocAucMulti(sigmoid=True, average='macro', sample_weight=None, max_fpr=None, log_metric=LogMetric.Valid, **kwargs):
    "Area Under the Receiver Operating Characteristic Curve for multi-label binary classification problems"
    activation = ActivationType.Sigmoid if sigmoid else ActivationType.No
    return skm_to_fastxtend(skm.roc_auc_score, activation=activation, flatten=False,
                         average=average, sample_weight=sample_weight, max_fpr=max_fpr, log_metric=log_metric, **kwargs)

# Internal Cell
def mse(inp,targ):
    "Mean squared error between `inp` and `targ`."
    return F.mse_loss(*flatten_check(inp,targ))

# Cell
def MSE(metric_type=MetricType.Avg, log_metric=LogMetric.Valid, **kwargs):
    "Mean squared error between `inp` and `targ`."
    return func_to_metric(mse, metric_type, False, log_metric=log_metric, **kwargs)

# Internal Cell
def rmse(inp, targ):
    return torch.sqrt(F.mse_loss(inp, targ))

# Cell
def RMSE(log_metric=LogMetric.Valid, **kwargs):
    "Mean squared error between `inp` and `targ`."
    return func_to_metric(rmse, MetricType.Accum, False, log_metric=log_metric, **kwargs)

# Internal Cell
def mae(inp,targ):
    "Mean absolute error between `inp` and `targ`."
    inp,targ = flatten_check(inp,targ)
    return torch.abs(inp - targ).mean()

# Cell
def MAE(metric_type=MetricType.Avg, log_metric=LogMetric.Valid, **kwargs):
    "Mean absolute error between `inp` and `targ`."
    return func_to_metric(mae, metric_type, False, log_metric=log_metric, **kwargs)

# Internal Cell
def msle(inp, targ):
    "Mean squared logarithmic error between `inp` and `targ`."
    inp,targ = flatten_check(inp,targ)
    return F.mse_loss(torch.log(1 + inp), torch.log(1 + targ))

# Cell
def MSLE(metric_type=MetricType.Avg, log_metric=LogMetric.Valid, **kwargs):
    "Mean squared logarithmic error between `inp` and `targ`."
    return func_to_metric(msle, metric_type, False, log_metric=log_metric, **kwargs)

# Internal Cell
def exp_rmspe(inp,targ):
    inp,targ = torch.exp(inp),torch.exp(targ)
    return torch.sqrt(((targ - inp)/targ).pow(2).mean())

# Cell
def ExpRMSE(log_metric=LogMetric.Valid, **kwargs):
    "Root mean square percentage error of the exponential of  predictions and targets"
    return func_to_metric(exp_rmspe, MetricType.Accum, False, log_metric=log_metric, **kwargs)

# Cell
def ExplainedVariance(sample_weight=None, log_metric=LogMetric.Valid, **kwargs):
    "Explained variance between predictions and targets"
    return skm_to_fastxtend(skm.explained_variance_score, is_class=False,
                         sample_weight=sample_weight, log_metric=log_metric, **kwargs)

# Cell
def R2Score(sample_weight=None, log_metric=LogMetric.Valid, **kwargs):
    "R2 score between predictions and targets"
    return skm_to_fastxtend(skm.r2_score, is_class=False, sample_weight=sample_weight,
                         log_metric=log_metric, **kwargs)

# Cell
def PearsonCorrCoef(dim_argmax=None, log_metric=LogMetric.Valid, **kwargs):
    "Pearson correlation coefficient for regression problem"
    def pearsonr(x,y): return scs.pearsonr(x,y)[0]
    return AccumMetricX(pearsonr, invert_arg=False, dim_argmax=dim_argmax,
                        log_metric=log_metric, **kwargs)

# Cell
def SpearmanCorrCoef(dim_argmax=None, axis=0, nan_policy='propagate', log_metric=LogMetric.Valid, **kwargs):
    "Spearman correlation coefficient for regression problem"
    def spearmanr(a,b=None,**kwargs): return scs.spearmanr(a,b,**kwargs)[0]
    return AccumMetricX(partial(spearmanr, axis=axis, nan_policy=nan_policy),
                        invert_arg=False, dim_argmax=dim_argmax, log_metric=log_metric, **kwargs)

# Internal Cell
def foreground_acc(inp, targ, bkg_idx=0, axis=1):
    "Computes non-background accuracy for multiclass segmentation"
    targ = cast(targ.squeeze(1), TensorBase)
    mask = targ != bkg_idx
    return (inp[mask]==targ[mask]).float().mean()

# Cell
def ForegroundAcc(bkg_idx=0, axis=1, metric_type=MetricType.Avg, log_metric=LogMetric.Valid, **kwargs):
    "Computes non-background accuracy for multiclass segmentation"
    return func_to_metric(foreground_acc, metric_type, True, bkg_idx=bkg_idx, axis=axis,
                          log_metric=log_metric, **kwargs)

# Cell
class Dice(MetricX):
    "Dice coefficient metric for binary target in segmentation"
    def __init__(self, axis=1, log_metric=LogMetric.Valid, **kwargs):
        super().__init__(dim_argmax=axis, log_metric=log_metric, **kwargs)
    def reset(self): self.inter,self.union = 0,0
    def accumulate(self, learn):
        super().accumulate(learn)
        self.pred,self.targ = flatten_check(self.pred, self.targ)
        self.inter += (self.pred*self.targ).float().sum().item()
        self.union += (self.pred+self.targ).float().sum().item()

    @property
    def value(self): return 2. * self.inter/self.union if self.union > 0 else None

# Cell
class DiceMulti(MetricX):
    "Averaged Dice metric (Macro F1) for multiclass target in segmentation"
    def __init__(self, axis=1, log_metric=LogMetric.Valid, **kwargs):
        super().__init__(dim_argmax=axis, log_metric=log_metric, **kwargs)
    def reset(self): self.inter,self.union = {},{}
    def accumulate(self, learn):
        super().accumulate(learn)
        self.pred,self.targ = flatten_check(self.pred, self.targ)
        for c in range(learn.pred.shape[self.dim_argmax]):
            p = torch.where(self.pred == c, 1, 0)
            t = torch.where(self.targ == c, 1, 0)
            c_inter = (p*t).float().sum().item()
            c_union = (p+t).float().sum().item()
            if c in self.inter:
                self.inter[c] += c_inter
                self.union[c] += c_union
            else:
                self.inter[c] = c_inter
                self.union[c] = c_union

    @property
    def value(self):
        binary_dice_scores = np.array([])
        for c in self.inter:
            binary_dice_scores = np.append(binary_dice_scores, 2.*self.inter[c]/self.union[c] if self.union[c] > 0 else np.nan)
        return np.nanmean(binary_dice_scores)

# Cell
class JaccardCoeff(Dice):
    "Implementation of the Jaccard coefficient that is lighter in RAM"
    @property
    def value(self): return self.inter/(self.union-self.inter) if self.union > 0 else None

# Cell
class CorpusBLEUMetric(MetricX):
    "BLEU Metric calculated over the validation corpus"
    def __init__(self, vocab_sz=5000, axis=-1, log_metric=LogMetric.Valid, name='CorpusBLEU', **kwargs):
        super().__init__(log_metric=log_metric, name=name, **kwargs)
        self.axis, self.vocab_sz = axis, vocab_sz
        self.pred_len,self.targ_len,self.samp_idx,self.corrects,self.counts, = 0,0,0,[0]*4,[0]*4

    def reset(self):
        self.pred_len,self.targ_len,self.corrects,self.counts = 0,0,[0]*4,[0]*4

    class NGram():
        def __init__(self, ngram, max_n=5000): self.ngram,self.max_n = ngram,max_n
        def __eq__(self, other):
            if len(self.ngram) != len(other.ngram): return False
            return np.all(np.array(self.ngram) == np.array(other.ngram))
        def __hash__(self): return int(sum([o * self.max_n**i for i,o in enumerate(self.ngram)]))

    def get_grams(self, x, n, max_n=5000):
        return x if n==1 else [self.NGram(x[i:i+n], max_n=max_n) for i in range(len(x)-n+1)]

    def get_correct_ngrams(self, pred, targ, n, max_n=5000):
        pred_grams,targ_grams = self.get_grams(pred, n, max_n=max_n),self.get_grams(targ, n, max_n=max_n)
        pred_cnt,targ_cnt = Counter(pred_grams),Counter(targ_grams)
        return sum([min(c, targ_cnt[g]) for g,c in pred_cnt.items()]),len(pred_grams)

    def accumulate(self, learn):
        if learn.training: return None
        else:
            last_output = learn.pred.argmax(dim=self.axis)
            last_target = learn.y
            for pred,targ in zip(last_output.cpu().numpy(),last_target.cpu().numpy()):
                self.pred_len += len(pred)
                self.targ_len += len(targ)
                smooth_mteval = 1
                for i in range(4):
                    c,t = self.get_correct_ngrams(pred, targ, i+1, max_n=self.vocab_sz)
                    if c == 0:
                        smooth_mteval *= 2
                        c = 1 / smooth_mteval    # exp smoothing, method 3 from http://acl2014.org/acl2014/W14-33/pdf/W14-3346.pdf
                    self.corrects[i] += c
                    self.counts[i]   += t

    @property
    def value(self):
        if self.counts == 0: return None
        elif max(self.corrects) == 0: return 0.0
        else:
            precs = [c/t for c,t in zip(self.corrects,self.counts)]
            len_penalty = math.exp(1 - self.targ_len/self.pred_len) if self.pred_len < self.targ_len else 1
            return len_penalty * ((precs[0]*precs[1]*precs[2]*precs[3]) ** 0.25)

# Cell
class Perplexity(AvgLossX):
    "Perplexity (exponential of cross-entropy loss) for Language Models"
    @property
    def value(self): return torch.exp(self.total/self.count) if self.count != 0 else None
    @property
    def name(self):  return "perplexity"

perplexity = Perplexity()

# Cell
class LossMetric(AvgMetricX):
    "Create a metric from `loss_func.attr` named `nm`"
    def __init__(self, attr, nm=None, log_metric=LogMetric.Valid, **kwargs):
        super().__init__(noop, log_metric=log_metric, **kwargs)
        store_attr('attr,nm')
    def accumulate(self, learn):
        bs = find_bs(learn.yb)
        self.total += learn.to_detach(getattr(learn.loss_func, self.attr, 0))*bs
        self.count += bs

    @property
    def name(self): return self.attr if self.nm is None else self.nm

# Cell
def LossMetrics(attrs, nms=None):
    "List of `LossMetric` for each of `attrs` and `nms`"
    if isinstance(attrs, str): attrs = attrs.split(',')
    nms = attrs if nms is None else nms.split(',') if isinstance(nms, str) else nms
    return [LossMetric(a, n) for a,n in zip(attrs,nms)]

# Internal Cell
try:
    import wandb
    from fastai.callback.wandb import WandbCallback
    from fastai.callback.training import FetchPredsCallback

    if not hasattr(WandbCallback,'_metrics_before_fit'): WandbCallback._metrics_before_fit = WandbCallback.before_fit

    @patch
    def before_fit(self:WandbCallback):
        self._metrics_before_fit()
        self.log_smooth = len(self.recorder.smooth_names) > 0

    @patch
    def after_batch(self:WandbCallback):
        "Log hyper-parameters and training loss"
        if self.training:
            self._wandb_step += 1
            self._wandb_epoch += 1/self.n_iter
            hypers = {f'{k}_{i}':v for i,h in enumerate(self.opt.hypers) for k,v in h.items()}
            if self.log_smooth:
                for n,m in zip(self.recorder.smooth_names, self.recorder.smooth_mets): hypers[n]=m.value
            wandb.log({'epoch': self._wandb_epoch, 'train_loss': to_detach(self.smooth_loss.clone()), 'raw_loss': to_detach(self.loss.clone()), **hypers}, step=self._wandb_step)

    @patch
    def after_epoch(self:WandbCallback):
        "Log validation loss and custom metrics & log prediction samples"
        # Correct any epoch rounding error and overwrite value
        self._wandb_epoch = round(self._wandb_epoch)
        wandb.log({'epoch': self._wandb_epoch}, step=self._wandb_step)
        # Log sample predictions
        if self.log_preds:
            try:
                self.log_predictions(self.learn.fetch_preds.preds)
            except Exception as e:
                self.log_preds = False
                self.remove_cb(FetchPredsCallback)
                print(f'WandbCallback was not able to get prediction samples -> {e}')
        wandb.log({n:s for n,s in zip(self.recorder.metric_names, self.recorder.log) if n not in ['train_loss', 'epoch', 'time']+self.recorder.smooth_names}, step=self._wandb_step)
except:
    pass

# Internal Cell
try:
    import tensorboard
    from fastai.callback.tensorboard import TensorBoardCallback, tensorboard_log

    if not hasattr(TensorBoardCallback,'_metrics_before_fit'): TensorBoardCallback._metrics_before_fit = TensorBoardCallback.before_fit

    @patch
    def before_fit(self:TensorBoardCallback):
        self._metrics_before_fit()
        self.log_smooth = len(self.recorder.smooth_names) > 0

    @patch
    def after_batch(self:TensorBoardCallback):
        self.writer.add_scalar('train_loss', self.smooth_loss, self.train_iter)
        for i,h in enumerate(self.opt.hypers):
            for k,v in h.items(): self.writer.add_scalar(f'{k}_{i}', v, self.train_iter)
        if self.log_smooth:
            for k,v in zip(self.recorder.smooth_names, self.recorder.smooth_mets):
                self.writer.add_scalar(f'{k}', v, self.train_iter)

    @patch
    def after_epoch(self:TensorBoardCallback):
        for n,v in zip(self.recorder.metric_names[2:-1], self.recorder.log[2:-1]):
            if n not in ['train_loss', 'time']+self.recorder.smooth_names:
                self.writer.add_scalar(n, v, self.train_iter)
        if self.log_preds:
            b = self.dls.valid.one_batch()
            self.learn.one_batch(0, b)
            preds = getattr(self.loss_func, 'activation', noop)(self.pred)
            out = getattr(self.loss_func, 'decodes', noop)(preds)
            x,y,its,outs = self.dls.valid.show_results(b, out, show=False, max_n=self.n_preds)
            tensorboard_log(x, y, its, outs, self.writer, self.train_iter)
except:
    pass


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/callback.cutmixup.ipynb (unless otherwise specified).

__all__ = ['CutMixUp', 'CutMixUpAugment']

# Cell
from fastai.basics import *
from fastai.data.transforms import IntToFloatTensor, Normalize
from fastai.callback.mixup import MixHandler, CutMix, MixUp, reduce_loss
from torch.distributions.beta import Beta

# Internal Cell
class CutMix(MixHandler):
    "Implementation of https://arxiv.org/abs/1905.04899"
    def __init__(self, alpha=1.): super().__init__(alpha)
    def before_batch(self):
        bs, _, H, W = self.x.size()
        self.lam = self.distrib.sample((1,)).to(self.x.device)
        shuffle = torch.randperm(bs).to(self.x.device)
        xb1,self.yb1 = self.x[shuffle], tuple((self.y[shuffle],))
        x1, y1, x2, y2 = self.rand_bbox(W, H, self.lam)
        self.learn.xb[0][..., y1:y2, x1:x2] = xb1[..., y1:y2, x1:x2]
        self.lam = (1 - ((x2-x1)*(y2-y1))/float(W*H))
        if not self.stack_y:
            ny_dims = len(self.y.size())
            self.learn.yb = tuple(L(self.yb1,self.yb).map_zip(torch.lerp,weight=unsqueeze(self.lam, n=ny_dims-1)))

    def rand_bbox(self, W, H, lam):
        cut_rat = torch.sqrt(1. - lam).to(self.x.device)
        cut_w = torch.round(W * cut_rat).type(torch.long).to(self.x.device)
        cut_h = torch.round(H * cut_rat).type(torch.long).to(self.x.device)
        # uniform
        cx = torch.randint(0, W, (1,)).to(self.x.device)
        cy = torch.randint(0, H, (1,)).to(self.x.device)
        x1 = torch.clamp(cx - torch.div(cut_w, 2, rounding_mode='floor'), 0, W)
        y1 = torch.clamp(cy - torch.div(cut_h, 2, rounding_mode='floor'), 0, H)
        x2 = torch.clamp(cx + torch.div(cut_w, 2, rounding_mode='floor'), 0, W)
        y2 = torch.clamp(cy + torch.div(cut_h, 2, rounding_mode='floor'), 0, H)
        return x1, y1, x2, y2

# Cell
class CutMixUp(MixUp, CutMix):
    "Combo implementation of https://arxiv.org/abs/1710.09412 and https://arxiv.org/abs/1905.04899"
    run_valid = False
    def __init__(self, mix_alpha=.4, cut_alpha=1., cutmix_ratio=1, mixup_ratio=1):
        MixUp.__init__(self, mix_alpha)
        CutMix.__init__(self, cut_alpha)
        self.mix_distrib = Beta(tensor(mix_alpha), tensor(mix_alpha))
        self.cut_distrib = Beta(tensor(cut_alpha), tensor(cut_alpha))
        self.ratio = mixup_ratio / (cutmix_ratio + mixup_ratio)

    def before_batch(self):
        if torch.rand(1) <= self.ratio: #mixup
            self.distrib = self.mix_distrib
            MixUp.before_batch(self)
        else:
            self.distrib = self.cut_distrib
            CutMix.before_batch(self)

# Cell
class CutMixUpAugment(MixUp, CutMix):
    "Combo implementation of https://arxiv.org/abs/1710.09412 and https://arxiv.org/abs/1905.04899 plus Augmentation"
    run_valid = False
    def __init__(self, mix_alpha=.4, cut_alpha=1., augment_ratio=1, cutmix_ratio=1, mixup_ratio=1, augs_only=None):
        MixUp.__init__(self, mix_alpha)
        CutMix.__init__(self, cut_alpha)
        self.mix_distrib = Beta(tensor(mix_alpha), tensor(mix_alpha))
        self.cut_distrib = Beta(tensor(cut_alpha), tensor(cut_alpha))
        self.aug_cutmix_ratio = augment_ratio / (augment_ratio + cutmix_ratio + mixup_ratio)
        if self.aug_cutmix_ratio == 1: self.cut_mix_ratio = 0
        else: self.cut_mix_ratio = mixup_ratio / (cutmix_ratio + mixup_ratio)
        self.augs_only = augs_only

    def before_fit(self):
        if self.augs_only is None: self.augs_only = (self.learn.n_epoch + 1)/self.learn.n_epoch
        elif self.augs_only >=1: self.augs_only = self.augs_only/self.learn.n_epoch
        else: self.augs_only = self.augs_only

        self._inttofloat_pipe = Pipeline([])
        self._norm_pipe = Pipeline([])

        # first copy transforms
        self._orig_pipe = self.dls.train.after_batch
        self._orig_pipe.split_idx = 0 # need to manually set split_idx for training augmentations to run

        # loop through existing transforms looking for IntToFloatTensor and Normalize
        for i in range(len(self.dls.train.after_batch.fs)):
            if isinstance(self.dls.train.after_batch[i], IntToFloatTensor):
                self._inttofloat_pipe = Pipeline([self.dls.train.after_batch[i]])
            elif isinstance(self.dls.train.after_batch[i], Normalize):
                self._norm_pipe = Pipeline([self.dls.train.after_batch[i]])

        # set existing transforms to an empty Pipeline
        self.dls.train.after_batch = Pipeline([])

    def before_batch(self):
        if self.augs_only >= self.learn.pct_train and torch.rand(1) >= self.aug_cutmix_ratio: # augs or mixup/cutmix
            self._aug = False
            self.learn.xb = self._inttofloat_pipe(self.xb) # apply inttofloat first
            if self.cut_mix_ratio > 0 and torch.rand(1) <= self.cut_mix_ratio: # mixup or cutmix
                self.distrib = self.mix_distrib
                MixUp.before_batch(self)
            else:
                self.distrib = self.cut_distrib
                CutMix.before_batch(self)
            self.learn.xb = self._norm_pipe(self.xb) # now normalize
        else:
            self._aug = True
            self.learn.xb = self._orig_pipe(self.xb) # original transforms

    def after_fit(self):
        self.dls.train.after_batch = self._orig_pipe

    def after_cancel_fit(self):
        self.after_fit()
        MixUp.after_cancel_fit(self)

    def lf(self, pred, *yb):
        if not self.training or self._aug: return self.old_lf(pred, *yb)
        with NoneReduce(self.old_lf) as lf:
            loss = torch.lerp(lf(pred,*self.yb1), lf(pred,*yb), self.lam)
        return reduce_loss(loss, getattr(self.old_lf, 'reduction', 'mean'))


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/audio.04_learner.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['StackSpecCallback', 'audio_learner']

# Cell
#nbdev_comment from __future__ import annotations

from fastcore.dispatch import retain_type

from fastai.callback.core import Callback
from fastai.callback.fp16 import MixedPrecision
from fastai.learner import Learner, defaults
from fastai.optimizer import Adam

from .core import TensorSpec, TensorMelSpec
from .data import MelSpectrogram, Spectrogram
from ..imports import *

# Cell
class StackSpecCallback(Callback):
    "Stacks tuples of TensorSpec or TensorMelSpec. ToDo: add resizing"
    order = MixedPrecision.order-1
    def before_batch(self):
        xb = L(self.xb)
        idx = xb.argwhere(lambda x: isinstance(x, (TensorSpec, TensorMelSpec)))
        ts = []
        for i in idx:
            ts.append(xb[i])
        stacked = torch.stack(ts, dim=2)
        xb = retain_type(torch.flatten(stacked, 1, 2), xb[i])
        self.learn.xb = tuple(xb)

# Cell
def audio_learner(
    dls,
    model,
    loss_func=None,
    opt_func=Adam,
    lr=defaults.lr,
    splitter=trainable_params,
    cbs=None,
    metrics=None,
    path=None,
    model_dir='models',
    wd=None,
    wd_bn_bias=False,
    train_bn=True,
    moms=(0.95,0.85,0.95)
) -> Learner:
    "An Audio specific Learner that stacks tuples of `TensorSpec` or `TensorMelSpec`"
    detuple = False
    for i in range(len(dls.train.after_batch.fs)):
        if not detuple and isinstance(dls.train.after_batch[i], (Spectrogram, MelSpectrogram)):
            detuple = is_listy(dls.train.after_batch[i].n_fft)

    if detuple:
        if cbs is None: cbs = DetupleSpecCallback()
        else: cbs = L(cbs) + L(DetupleSpecCallback())

    return Learner(dls=dls, model=model, loss_func=loss_func, opt_func=opt_func, lr=lr, splitter=splitter, cbs=cbs,
                    metrics=metrics, path=path, model_dir=model_dir, wd=wd, wd_bn_bias=wd_bn_bias, train_bn=train_bn,
                    moms=moms)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/vision.models.pooling.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['BlurPool', 'MaxBlurPool']

# Cell
#nbdev_comment from __future__ import annotations

from kornia.filters import BlurPool2D, MaxBlurPool2D

from ...imports import *

# Cell
def BlurPool(
    stride:int=2, # The stride size for pooling
    ks:int=3, # The kernel size for pooling
    padding:int=0, # Unused, for fastai compatibility
    ndim:int=2, # Unused, for fastai compatibility
    ceil_mode:bool=False # Unused, for fastai compatibility
) -> BlurPool2D:
    "Compute blur (anti-aliasing) and downsample a given feature map."
    assert ndim==2, "BlurPool is only implemented for 2D"
    return BlurPool2D(kernel_size=ks, stride=stride)

# Cell
def MaxBlurPool(
    stride:int=2, # The stride size for blur pooling
    ks:int=3, # The kernel size for blur pooling
    padding:int=0, # Unused, for fastai compatibility
    ndim:int=2, # Unused, for fastai compatibility
    ceil_mode:int=True, # If True, output size matches conv2d with same kernel size
    max_ks:int=2 # The kernel size for max pooling
) -> MaxBlurPool2D:
    "Compute pools and blurs and downsample a given feature map. Equivalent to `nn.Sequential(nn.MaxPool2d(...), BlurPool2D(...))`"
    assert ndim==2, "MaxBlurPool is only implemented for 2D"
    return MaxBlurPool2D(kernel_size=ks, stride=stride, ceil_mode=ceil_mode, max_pool_size=max_ks)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/callback.ema.ipynb (unless otherwise specified).

__all__ = ['EMACallback']

# Cell
try:
    import timm
except ImportError:
    raise ImportError("timm is required to use EMACallback. Install via `pip install timm`.")

from timm.utils.model_ema import ModelEmaV2

from fastai.callback.core import Callback
from fastai.callback.fp16 import MixedPrecision

from ..imports import *

# Cell
class EMACallback(Callback):
    order,run_valid = MixedPrecision.order+1,False
    "Callback to implment Model Exponential Moving Average from PyTorch Image Models in fast.ai"
    def __init__(self, decay=0.9998, ema_device=None):
        store_attr()

    @torch.no_grad()
    def before_fit(self):
        self.ema_model = ModelEmaV2(self.learn.model, self.decay, self.ema_device)

    def after_batch(self):
        self.ema_model.update(self.learn.model)

    def before_validate(self):
        self.temp_model = self.learn.model
        self.learn.model = self.ema_model.module

    def after_validate(self):
        self.learn.model = self.temp_model

    @torch.no_grad()
    def after_fit(self):
        self.learn.model = self.ema_model.module
        self.ema_model = None
        self.remove_cb(EMACallback)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/0_basics.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['is_listish', 'listify_store_attr']

# Cell
#nbdev_comment from __future__ import annotations

import sys
import re

from fastcore.basics import annotations, argnames, _store_attr, filter_dict, range_of, Inf
from fastcore.dispatch import typedispatch, retain_meta

from .imports import *

# Cell
def is_listish(x):
    "Subset of `is_listy`: (tuple,list,L)"
    return isinstance(x, (tuple,list,L))

# Cell
def listify_store_attr(names=None, self=None, but='', cast=False, store_args=None, **attrs):
    "Maybe listify, then store params named in comma-separated `names` from calling context into attrs in `self`"
    fr = sys._getframe(1)
    args = argnames(fr, True)
    if self: args = ('self', *args)
    else: self = fr.f_locals[args[0]]
    if store_args is None: store_args = not hasattr(self,'__slots__')
    if store_args and not hasattr(self, '__stored_args__'): self.__stored_args__ = {}
    anno = annotations(self) if cast else {}
    if names and isinstance(names,str): names = re.split(', *', names)
    ns = names if names is not None else getattr(self, '__slots__', args[1:])
    added = {n:fr.f_locals[n] for n in ns}
    attrs = {**attrs, **added}
    if isinstance(but,str): but = re.split(', *', but)
    # start listify_store_attr code
    attrs = {k:v for k,v in attrs.items() if k not in but}
    attrs_len = {n:len(attrs[n]) if is_listy(attrs[n]) else 1 for n in attrs.keys()}
    l = max(attrs_len.values())
    if l > 1:
        ones = filter_dict(attrs_len, lambda x,v: v==1)
        if len(ones)+1 != len(attrs_len):
            raise ValueError(f'Args must be all be length {l} or 1. Invalid args: {list(filter_dict(attrs_len, lambda x,v: l>v>1).keys())}')
        for n in ones.keys():
            # if-else needed for None input
            attrs[n] = L(attrs[n])*l if is_listy(attrs[n]) else L([attrs[n]])*l
    return _store_attr(self, anno, **attrs)

# Cell
@typedispatch
def show_batch(x, y, samples, ctxs=None, max_n=9, **kwargs):
    if ctxs is None: ctxs = Inf.nones
    plots = []
    if hasattr(samples[0], 'show'):
        for s,c,_ in zip(samples,ctxs,range(max_n)):
            s = retain_meta(x, s)
            plots.append(s.show(ctx=c, **kwargs))
    else:
        for i in range_of(samples[0]):
            for b,c,_ in zip(samples.itemgot(i),ctxs,range(max_n)):
                b = retain_meta(x, b)
                plots.append(b.show(ctx=c, **kwargs))
    ctxs = plots
    return ctxs


import math
from collections import OrderedDict, defaultdict, Counter, namedtuple
from enum import Enum
from functools import partial
from typing import Iterable, Union, TypeVar, Callable, Any
from warnings import warn

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from fastcore.basics import store_attr, noop, Self, hasattrs, partialler
from fastcore.foundation import L, fastuple, patch
from fastcore.meta import delegates
from fastcore.utils import ifnone
from fastcore.xtras import is_listy

from fastai.torch_core import *

T = TypeVar('T')

listy     = Union[Iterable[T], L, fastuple]
listified = Union[T, Iterable[T], L, fastuple]
Number    = Union[int, float]


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/callback.lr_finder.ipynb (unless otherwise specified).

__all__ = ['LRFinder']

# Cell
from fastcore.xtras import is_listy
from fastcore.foundation import patch, docs, Path
from fastcore.basics import tuplify
from fastai.callback.schedule import ParamScheduler, SchedExp, SuggestionMethod
from fastai.torch_core import tensor, get_random_states, set_random_states
from fastai.learner import Learner, CancelFitException, CancelValidException
from functools import partial
from copy import deepcopy
import torch
import collections, tempfile

# Cell
@docs
class LRFinder(ParamScheduler):
    "Training with exponentially growing learning rate"
    def __init__(self, start_lr=1e-7, end_lr=10, num_it=100, stop_div=True, restore_state=True):
        if num_it < 6: num_it = 6
        self.scheds = {'lr': [SchedExp(s, e) for (s,e) in zip(start_lr,end_lr)
                             ] if is_listy(start_lr) else SchedExp(start_lr, end_lr)}
        self.num_it,self.stop_div,self.restore_state = num_it,stop_div,restore_state

    def before_fit(self):
        super().before_fit()
        if self.restore_state:
            self.old_dls = deepcopy(self.learn.dls)
            self.states = get_random_states()
        path = self.path/self.model_dir
        path.mkdir(parents=True, exist_ok=True)
        self.tmp_d = tempfile.TemporaryDirectory(dir=path)
        self.tmp_p = Path(self.tmp_d.name).stem
        self.learn.save(f'{self.tmp_p}/_tmp')
        self.best_loss = float('inf')

    def before_batch(self): self._update_val(self.train_iter/self.num_it)

    def after_batch(self):
        super().after_batch()
        if self.smooth_loss < self.best_loss: self.best_loss = self.smooth_loss
        if self.smooth_loss > 4*self.best_loss and self.stop_div: raise CancelFitException()
        if self.train_iter >= self.num_it: raise CancelFitException()

    def before_validate(self): raise CancelValidException()

    def after_fit(self):
        self.learn.opt.zero_grad() # Needed before detaching the optimizer for future fits
        tmp_f = self.path/self.model_dir/self.tmp_p/'_tmp.pth'
        if tmp_f.exists():
            self.learn.load(f'{self.tmp_p}/_tmp', with_opt=True)
            self.tmp_d.cleanup()
        if self.restore_state:
            self.learn.dls = self.old_dls
            set_random_states(**self.states)

    _docs = {"before_fit": "Initialize container for hyper-parameters and save the model & optimizer, optionally saving dataloader & random state",
             "before_batch": "Set the proper hyper-parameters in the optimizer",
             "after_batch": "Record hyper-parameters of this batch and potentially stop training",
             "after_fit": "Save the hyper-parameters in the recorder if there is one and load the original model & optimizer, optionally restoring dataloader & random state",
             "before_validate": "Skip the validation part of training"}

# Cell
@patch
def lr_find(self:Learner, start_lr=1e-7, end_lr=10, num_it=100, stop_div=True, show_plot=True, suggest_funcs=(SuggestionMethod.Valley), restore_state=True):
    "Launch a mock training to find a good learning rate and return suggestions based on `suggest_funcs` as a named tuple. Use `restore_state` to reset dataloaders and random state after running."
    n_epoch = num_it//len(self.dls.train) + 1
    cb=LRFinder(start_lr=start_lr, end_lr=end_lr, num_it=num_it, stop_div=stop_div,restore_state=restore_state)
    with self.no_logging(): self.fit(n_epoch, cbs=cb)
    if suggest_funcs is not None:
        lrs, losses = tensor(self.recorder.lrs[num_it//10:-5]), tensor(self.recorder.losses[num_it//10:-5])
        nan_idxs = torch.nonzero(torch.isnan(losses.view(-1)))
        if len(nan_idxs) > 0:
            drop_idx = min(nan_idxs)
            lrs = lrs[:drop_idx]
            losses = losses[:drop_idx]
        _suggestions, nms = [], []
        for func in tuplify(suggest_funcs):
            nms.append(func.__name__ if not isinstance(func, partial) else func.func.__name__) # deal with partials
            _suggestions.append(func(lrs, losses, num_it))

        SuggestedLRs = collections.namedtuple('SuggestedLRs', nms)
        lrs, pnts = [], []
        for lr, pnt in _suggestions:
            lrs.append(lr)
            pnts.append(pnt)
        if show_plot: self.recorder.plot_lr_find(suggestions=pnts, nms=nms)
        return SuggestedLRs(*lrs)

    elif show_plot: self.recorder.plot_lr_find()


from .core import *
from .data import *
from .augment import *
from .learner import *
from .mixup import *
from ..basics import *
from ..callback import channelslast, lr_finder, casttotensor, tracker
from ..data.all import *
from ..losses import *
from ..metrics import *
from ..multiloss import *
from ..utils import *


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/audio.03_augment.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['Flip', 'Roll', 'AudioPadMode', 'RandomCropPad', 'VolumeMode', 'Volume', 'PeakNorm', 'VolumeOrPeakNorm',
           'Noise', 'VolumeBatch', 'PitchShift', 'PitchShiftTA', 'TimeStretch', 'PitchShiftOrTimeStretch',
           'TimeMasking', 'FrequencyMasking', 'AmplitudeToDBMode', 'AmplitudeToDB', 'AudioNormalize']

# Cell
#nbdev_comment from __future__ import annotations

import random
import math
from collections import Counter
from fractions import Fraction
from functools import reduce
from itertools import chain, count, islice, repeat

import colorednoise
from primePy import primes

from torch import _VF
from torch.distributions import Bernoulli
import torchaudio.transforms as tatfms
import torchaudio.functional as TAF
from torchaudio.functional.functional import _get_sinc_resample_kernel, _apply_sinc_resample_kernel

from fastcore.transform import DisplayedTransform, retain_type

from fastai.data.load import DataLoader
from fastai.vision.augment import RandTransform, broadcast_vec

from ..transform import BatchRandTransform
from .core import TensorAudio, TensorSpec, TensorMelSpec
from ..imports import *

# Internal Cell
def _get_audio_attr(x:TensorAudio|TensorSpec|TensorMelSpec|tuple[TensorAudio|TensorSpec|TensorMelSpec,...], attr:str) -> Any:
    if isinstance(x, tuple): return getattr(x[0], attr)
    else:                    return getattr(x, attr)

# Cell
class Flip(RandTransform):
    "Randomly flip `TensorAudio` with probability `p`"
    order, split_idx = 5, 0
    def __init__(self, p:float=0.5):
        super().__init__(p=p)

    def encodes(self, x:TensorAudio) -> Tensor:
        return x.fliplr()

# Cell
class Roll(RandTransform):
    "Randomly shift `TensorAudio` with rollover"
    order, split_idx = 6, 0
    def __init__(self,
        p:float=0.5, # Per-item probability
        max_roll:float=0.5, # Maximum shift
    ):
        store_attr('max_roll')
        super().__init__(p=p)

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        super().before_call(b, split_idx)
        if split_idx==0: self.shift = int(random.uniform(-1,1) * self.max_roll * _get_audio_attr(b, 'samples'))

    def encodes(self, x:TensorAudio) -> Tensor:
        return x.roll(shifts=self.shift, dims=-1)

# Cell
class AudioPadMode(Enum):
    "All methods of padding audio as attributes to get tab-completion and typo-proofing",
    Constant = 1
    ConstantPre = 2
    ConstantPost = 3
    Repeat = 4

# Internal Cell
@patch
def crop_pad(x:TensorAudio,
    end_len:int,
    crop_start:int|None=None,
    pad_len:int=0,
    padmode:AudioPadMode=AudioPadMode.Repeat,
    constant:float=0
):
    if crop_start is not None:
        x = x[...,crop_start:crop_start+end_len]
    elif pad_len:
        if padmode==AudioPadMode.ConstantPre:
            x = _VF.constant_pad_nd(x, (pad_len, 0), constant)
        elif padmode==AudioPadMode.ConstantPost:
            x = _VF.constant_pad_nd(x, (0, pad_len), constant)
        elif padmode==AudioPadMode.Repeat:
            if pad_len >= x.samples:
                x = x.repeat(1, 2+max(pad_len//x.samples, 2))
            else:
                x = torch.cat([x, x[...,0:pad_len]], dim=-1)
        elif padmode==AudioPadMode.Constant:
            if pad_len % 2 == 0: pad = pad_len//2
            else:                pad = pad_len//2 + 1
            x = _VF.constant_pad_nd(x, (pad, pad), constant)

    if x.samples > end_len:
        x = x[...,0:end_len]
    return x

# Cell
class RandomCropPad(RandTransform):
    "Randomly resize `TensorAudio` to specified length. Center crop during valid."
    split_idx, order = None, 25
    def __init__(self,
        duration:Number|None=None, # Crop length in seconds
        samples:int|None=None, # Crop length in samples
        padmode:AudioPadMode=AudioPadMode.Repeat, # How to pad if necessary
        constant:Number=0, # Value for `AudioPadMode.Constant`
    ):
        if duration is None and samples is None:
            raise ValueError('One of `duration` or `samples` must be set')
        if constant >= 1 or constant <= -1:
            raise ValueError('`constant` must be between [-1,1]')
        if padmode not in [vol for vol in AudioPadMode]:
            raise ValueError(f"`padmode` {padmode} is not valid")
        store_attr()
        super().__init__()

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        self.samples = self.samples if self.samples is not None else int(self.duration*_get_audio_attr(b, 'sr'))
        self.orig_samples = _get_audio_attr(b, 'samples')
        if split_idx:
            self.crop_start = (self.orig_samples-self.samples)//2 if self.samples < self.orig_samples else None
        else:
            self.crop_start = torch.randint(0,self.orig_samples-self.samples, (1,)) if self.samples < self.orig_samples else None

        self.pad_len = (self.samples-self.orig_samples) if self.samples > self.orig_samples else 0

    def encodes(self, x:TensorAudio) -> Tensor:
        return x.crop_pad(self.samples, self.crop_start, self.pad_len, self.padmode, self.constant)

# Cell
class VolumeMode(Enum):
    "All Volume modes as attributes to get tab-completion and typo-proofing",
    DB = 1
    Power = 2
    Amplitude = 3

# Cell
class Volume(RandTransform):
    "Randomly change `TensorAudio`'s volume"
    order, split_idx = 30, 0
    def __init__(self,
        p:float=0.75, # Per-item probability
        gain:Number|None=None, # Gain is a positive amplitude ratio, a power (voltage squared), or in decibels. If none, randomly select from `gain_range`.
        gain_range:tuple[Number,Number]=(-18, 6), # Random range for gain
        volmode:VolumeMode=VolumeMode.DB # One of `VolumeMode.DB`, `VolumeMode.Amplitude`, or `VolumeMode.Power`
    ):
        super().__init__(p=p)
        store_attr(but='p')
        if volmode not in [vol for vol in VolumeMode]:
            raise ValueError(f"`vol_mode` {volmode} is not valid")
        if volmode != VolumeMode.DB:
            self.gain_range = (max(gain_range[0],0), gain_range[1])
        self._random_gain = gain is None

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        super().before_call(b, split_idx)
        if split_idx==0 and self._random_gain: self.gain = random.uniform(*self.gain_range)

    def encodes(self, x:TensorAudio) -> Tensor:
        if self.volmode == VolumeMode.DB:
            x = TAF.gain(x, self.gain)
        elif self.volmode == VolumeMode.Amplitude:
            x = x * self.gain
        elif self.volmode == VolumeMode.Power:
            x = TAF.gain(x, 10 * math.log10(self.gain))

        return torch.clamp(x, -1, 1)

# Cell
class PeakNorm(RandTransform):
    "Randomly apply constant gain so `TensorAudio`'s loudest level is between -1 and 1 with probability `p`"
    order, split_idx = 31, 0
    def __init__(self, p:float=0.1):
        super().__init__(p=p)
        store_attr(but='p')

    def encodes(self, x:TensorAudio) -> Tensor:
            return x / torch.max(torch.abs(x))

# Cell
class VolumeOrPeakNorm(RandTransform):
    "Randomly apply `Volume` or `Peak` to `TensorAudio` in one transform"
    order, split_idx = 30, 0
    def __init__(self,
        p:float=0.75, # Per-item probability
        peak_p:float=0.1, # Probability of applying `Peak`
        gain:Number|None=None, # Gain is a positive amplitude ratio, a power (voltage squared), or in decibels. If none, randomly select from `gain_range`.
        gain_range:tuple[Number,Number]=(-18, 6), # Random range for gain
        volmode:VolumeMode=VolumeMode.DB # One of `VolumeMode.DB`, `VolumeMode.Amplitude`, or `VolumeMode.Power`
    ):
        super().__init__(p=p)
        store_attr(but='p')
        if volmode not in [vol for vol in VolumeMode]:
            raise ValueError(f"`vol_mode` {volmode} is not valid")
        if volmode != VolumeMode.DB:
            self.gain_range = (max(gain_range[0],0), gain_range[1])
        self._random_gain = gain is None

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        super().before_call(b, split_idx)
        if split_idx==0:
            self.peak = torch.randn(1) < self.peak_p
            if not self.peak and self._random_gain:
                self.gain = random.uniform(*self.gain_range)

    def encodes(self, x:TensorAudio) -> Tensor:
        if self.peak:
            x = x / torch.max(torch.abs(x))
        else:
            if self.volmode == VolumeMode.DB:
                x = TAF.gain(x, self.gain)
            elif self.volmode == VolumeMode.Amplitude:
                x = x * self.gain
            elif self.volmode == VolumeMode.Power:
                x = TAF.gain(x, 10 * math.log10(self.gain))

        return torch.clamp(x, -1, 1)

# Internal Cell

# Noise contains code from github.com/fastaudio/fastaudio

# MIT License

# Original FastAudio implementation, Copyright (c) 2020 scart97

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

class NoiseColor(Enum):
    """Exponent beta values for named noise colors (see `colored_noise`)."""
    Violet = -2
    Blue = -1
    White = 0
    Pink = 1
    Brown = 2

# Cell
class Noise(RandTransform):
    "Adds noise of specified color and level to `TensorAudio` relative to mean audio level"
    order, split_idx = 60, 0
    def __init__(self,
        p=0.25, # Per-item probability
        noise_level:float|None=None, # Loudness of noise, if None randomly selects from `noise_range`
        noise_range:tuple[float,float]=(0.01,0.1), # Range of noise loudness values
        color:NoiseColor|None=None, # Color of noise to add, if None randomly selects from `NoiseColor`
    ):
        super().__init__(p=p)
        store_attr(but='p')
        self._color_choice = [nc for nc in NoiseColor]
        if color is not None and color not in self._color_choice:
            raise ValueError(f"`color` {color} is not valid")
        self._random_color = color is None
        self._random_noise = noise_level is None

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        super().before_call(b, split_idx)
        if split_idx==0:
            if self._random_color: self.color = random.choice(self._color_choice)
            if self._random_noise: self.noise_level = random.uniform(*self.noise_range)

    def encodes(self, x:TensorAudio) -> Tensor:
        # if it's white noise, use torch.randn_like for speed
        if self.color == NoiseColor.White:
            noise = torch.randn_like(x)
        else:
            noise = torch.from_numpy(colorednoise.powerlaw_psd_gaussian(exponent=self.color.value, size=x.samples)).float()
        x += noise * x.abs().mean() * self.noise_level
        return x

# Cell
class VolumeBatch(BatchRandTransform):
    "Randomly change `TensorAudio`'s volume on the GPU"
    order, split_idx = 10, 0
    def __init__(self,
        p:float=0.5, # Per-item probability
        gain:Number|None=None, # Gain is a positive amplitude ratio, a power (voltage squared), or in decibels. If none, randomly select from `gain_range`.
        gain_range:tuple[Number,Number]=(-18, 6), # Random range for gain
        volmode:VolumeMode=VolumeMode.DB # One of `VolumeMode.DB`, `VolumeMode.Amplitude`, or `VolumeMode.Power`
    ):
        super().__init__(p=p)
        store_attr(but='p')
        if volmode not in [vol for vol in VolumeMode]:
            raise ValueError(f"`vol_mode` {volmode} is not valid")
        if volmode != VolumeMode.DB:
            self.gain_range = (max(gain_range[0],0), gain_range[1])
        self._random_gain = gain is None

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        super().before_call(b, split_idx)
        if split_idx==0 and self._random_gain: self.gain = random.uniform(*self.gain_range)

    def encodes(self, x:TensorAudio) -> Tensor:
        if self.volmode == VolumeMode.DB:
            x = TAF.gain(x, self.gain)
        elif self.volmode == VolumeMode.Amplitude:
            x = x * self.gain
        elif self.volmode == VolumeMode.Power:
            x = TAF.gain(x, 10 * math.log10(self.gain))

        return torch.clamp(x, -1, 1)

# Internal Cell
# PitchShift & TimeStretch contains code from https://github.com/KentoNishi/torch-pitch-shift and https://github.com/KentoNishi/torch-time-stretch, respectively

# MIT License

# Original implementation Copyright (c) 2021 Kento Nishi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Internal Cell

# https://stackoverflow.com/a/46623112/9325832
def _combinations_without_repetition(r, iterable=None, values=None, counts=None):
    if iterable:
        values, counts = zip(*Counter(iterable).items())

    f = lambda i, c: chain.from_iterable(map(repeat, i, c))
    n = len(counts)
    indices = list(islice(f(count(), counts), r))
    if len(indices) < r:
        return
    while True:
        yield tuple(values[i] for i in indices)
        for i, j in zip(reversed(range(r)), f(reversed(range(n)), reversed(counts))):
            if indices[i] != j:
                break
        else:
            return
        j = indices[i] + 1
        for i, j in zip(range(i, r), f(count(j), counts[j:])):
            indices[i] = j

# Internal Cell
def _get_fast_shifts(
    sample_rate:int, # Audio sample rate
    condition:Callable[...,bool]=lambda x: x >= 0.5 and x <= 2 and x != 1, # Function to validate fast shift ratios
) -> list[Fraction]: # Fast pitch-shift target ratios
    "Search for pitch-shift targets that can be computed quickly for a given sample rate"
    fast_shifts = set()
    factors = primes.factors(sample_rate)
    products = []
    for i in range(1, len(factors) + 1):
        products.extend(
            [
                reduce(lambda x, y: x * y, x)
                for x in _combinations_without_repetition(i, iterable=factors)
            ]
        )
    for i in products:
        for j in products:
            f = Fraction(i, j)
            if condition(f):
                fast_shifts.add(f)
    return list(fast_shifts)

# Internal Cell
def _semitones_to_ratio(semitones:float) -> Fraction:
    "Convert semitonal shifts into a Fraction indicating a pitch shift ratio"
    return Fraction(2.0 ** (semitones / 12.0))

# Internal Cell
def _get_fast_stretches(
    sample_rate:int, # Audio sample rate
    condition:Callable[...,bool]=lambda x: x >= 0.5 and x <= 2 and x != 1, # Function to validate fast stretch ratios.
) -> list[Fraction]: # Fast time-stretch target ratios
    "Search for time-stretch targets that can be computed quickly for a given sample rate"
    fast_shifts = set()
    factors = primes.factors(sample_rate)
    products = []
    for i in range(1, len(factors) + 1):
        products.extend(
            [
                reduce(lambda x, y: x * y, x)
                for x in _combinations_without_repetition(i, iterable=factors)
            ]
        )
    for i in products:
        for j in products:
            f = Fraction(i, j)
            if condition(f):
                fast_shifts.add(f)
    return list(fast_shifts)

# Internal Cell
@patch
def pitch_shift(x:TensorAudio, n_fft, hop_length, shift, sr, new_sr, gcd, kernel, width, padmode, constant, device):
    shape = x.shape
    x = x.reshape(shape[0] * shape[1], shape[2])
    x = torch.stft(x, n_fft, hop_length, return_complex=True)
    phase_advance = torch.linspace(0, math.pi * hop_length, x.shape[1], device=device)[..., None]
    x = TAF.phase_vocoder(x, float(1 / shift), phase_advance)
    phase_advance = None
    x = torch.istft(x, n_fft, hop_length)
    x = retain_type(_apply_sinc_resample_kernel(x, sr, new_sr, gcd, kernel, width), typ=TensorAudio)
    crop_start = torch.randint(0, x.shape[-1]-shape[-1], (1,)) if shape[-1] < x.shape[-1]  else None
    pad_len = (shape[-1]-x.shape[-1]) if shape[-1] > x.shape[-1] else 0
    return x.crop_pad(shape[-1], crop_start, pad_len, padmode, constant).reshape(*shape)

# Cell
class PitchShift(BatchRandTransform):
    "Fast shift of `TensorAudio`'s pitch."
    order, split_idx = 20, 0
    def __init__(self,
        p:float=0.2, # Per-item probability
        semitones:tuple[float,float]=(-4.0,4.0), # Random pitch shift range in semitones to compute efficient shifts
        bins_per_octave:int=12, # Number of steps per octave
        padmode:AudioPadMode=AudioPadMode.Repeat, # How to pad if necessary
        constant:Number=0, # Value for `AudioPadMode.Constant`
        split:int|None=None # Apply transform to `split` items at a time. Use to prevent GPU OOM.
    ):
        super().__init__(p=p)
        store_attr(but='p')
        self.sr = 0
        if padmode not in [vol for vol in AudioPadMode]:
            raise ValueError(f"`padmode` {padmode} is not valid")

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        super().before_call(b, split_idx)
        if not self.sr:
            self.sr = _get_audio_attr(b, 'sr')
            self.n_fft = self.sr // 64
            self.hop_length = self.n_fft // 32
            self._fast_shifts = _get_fast_shifts(self.sr, self._validate_fast_shifts)
            self.type = _get_audio_attr(b, 'dtype')
            self.device = _get_audio_attr(b, 'device')
        if split_idx==0:
            self.shift = random.choice(self._fast_shifts)
            if not isinstance(self.shift, Fraction):
                self.shift = 2.0 ** (float(self.shift) / self.bins_per_octave)
            self.new_sr = int(self.sr/self.shift)
            self.gcd = math.gcd(self.sr, self.new_sr)
            self.kernel, self.width = _get_sinc_resample_kernel(self.sr, self.new_sr, self.gcd,
                                                                6, 0.99, 'sinc_interpolation',
                                                                None, self.device, self.type)

    def encodes(self, x:TensorAudio) -> Tensor:
        if self.split:
            xs = []
            for xi in retain_type(x.split(self.split), TensorAudio):
                xs.append(xi.pitch_shift(self.n_fft, self.hop_length, self.shift, self.sr, self.new_sr,
                                         self.gcd, self.kernel, self.width, self.padmode, self.constant, self.device))
            return torch.cat(xs, dim=0)
        else:
            return x.pitch_shift(self.n_fft, self.hop_length, self.shift, self.sr, self.new_sr,
                                 self.gcd, self.kernel, self.width, self.padmode, self.constant, self.device)

    def _validate_fast_shifts(self, x):
        return x >= _semitones_to_ratio(self.semitones[0]) and x <= _semitones_to_ratio(self.semitones[1]) and x != 1

# Cell
class PitchShiftTA(BatchRandTransform):
    "Shift the `TensorAudio`'s pitch using TorchAudio. Can be slower than `PitchShift`"
    order, split_idx = 20, 0
    def __init__(self,
        p:float=0.2, # Per-item probability
        n_steps:int|None=None, # The (fractional) steps to shift waveform
        n_step_range:tuple[int,int] = (2,6), # Random `n_steps` range if `n_steps` is None
        bins_per_octave:int=12, # Number of steps per octave
        n_fft:int=512, # Size of FFT, creates n_fft // 2 + 1 bins
        win_length:int|None=None, # Window size. Defaults to `n_fft`
        hop_length:int|None=None, # Length of hop between STFT windows. Defaults to `win_length` // 4
        window_fn:Callable[..., Tensor]=torch.hann_window, # Window tensor applied/multiplied to each frame/window
        wkwargs:dict|None=None, # Args for `window_fn`
    ):
        super().__init__(p=p)
        store_attr(but='p,window_fn,wkwargs')
        self.sr=0
        self.win_length = win_length if win_length is not None else n_fft
        self.hop_length = hop_length if hop_length is not None else self.win_length // 4
        self.window = window_fn(self.win_length) if wkwargs is None else window_fn(self.win_length, **wkwargs)

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        super().before_call(b, split_idx)
        if not self.sr: self.sr = _get_audio_attr(b, 'sr')
        if split_idx==0:
            self.steps = random.uniform(*self.n_step_range) if self.n_steps is None else self.n_steps

    def encodes(self, x:TensorAudio) -> Tensor:
        return TAF.pitch_shift(x, self.sr, self.steps, self.bins_per_octave, self.n_fft,
                               self.win_length, self.hop_length, self.window)

# Internal Cell
@patch
def time_stretch(x:TensorAudio, n_fft, hop_length, stretch, padmode, constant, device):
    shape = x.shape
    x = x.reshape(shape[0] * shape[1], shape[2])
    x = torch.stft(x, n_fft, hop_length, return_complex=True)
    phase_advance = torch.linspace(0, math.pi * hop_length, x.shape[1], device=device)[..., None]
    x = TAF.phase_vocoder(x, float(1/stretch), phase_advance)
    phase_advance = None
    x = torch.istft(x, n_fft, hop_length)
    crop_start = torch.randint(0, x.shape[-1]-shape[-1], (1,)) if shape[-1] < x.shape[-1]  else None
    pad_len = (shape[-1]-x.shape[-1]) if shape[-1] > x.shape[-1] else 0
    return x.crop_pad(shape[-1], crop_start, pad_len, padmode, constant).reshape(*shape)

# Cell
class TimeStretch(BatchRandTransform):
    "Fast time stretch of `TensorAudio`"
    order, split_idx = 25, 0
    def __init__(self,
        p:float=0.2, # Per-item probability
        stretch_rates:tuple[float,float]=(0.5,2.0), # Random time stretch range to compute efficient stretches. Defaults to 50%-200% speed
        padmode:AudioPadMode=AudioPadMode.Repeat, # How to pad if necessary
        constant:Number=0, # Value for `AudioPadMode.Constant`
        split:int|None=None # Apply transform to `split` items at a time. Use to prevent GPU OOM.
    ):
        super().__init__(p=p)
        store_attr(but='p')
        self.sr = 0
        if padmode not in [vol for vol in AudioPadMode]:
            raise ValueError(f"`padmode` {padmode} is not valid")

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        super().before_call(b, split_idx)
        if not self.sr:
            self.sr = _get_audio_attr(b, 'sr')
            self.n_fft = self.sr // 64
            self.hop_length = self.n_fft // 32
            self._fast_stretchs = _get_fast_stretches(self.sr, self._validate_fast_stretches)
            self.device = _get_audio_attr(b, 'device')
        if split_idx==0:
            self.stretch = float(1/random.choice(self._fast_stretchs))

    def encodes(self, x:TensorAudio) -> Tensor:
        if self.split:
            xs = []
            for xi in retain_type(x.split(self.split), TensorAudio):
                xs.append(xi.time_stretch(self.n_fft, self.hop_length, self.stretch, self.padmode, self.constant, self.device))
            return torch.cat(xs, dim=0)
        else:
            return x.time_stretch(self.n_fft, self.hop_length, self.stretch, self.padmode, self.constant, self.device)


    def _validate_fast_stretches(self, x):
        return x >= self.stretch_rates[0] and x <= self.stretch_rates[1] and x != 1

# Cell
class PitchShiftOrTimeStretch(BatchRandTransform):
    "Randomly apply either `PitchShift` or `TimeStretch` `TensorAudio` to minimize distortion"
    order, split_idx = 25, 0
    def __init__(self,
        p:float=0.4, # Per-item probability
        semitones:tuple[float,float]=(-4.0,4.0), # Random pitch shift range in semitones to compute efficient shifts
        bins_per_octave:int=12, # Number of steps per octave
        stretch_rates:tuple[float,float]=(0.5,2.0), # Random time stretch range to compute efficient stretches. Defaults to 50%-200% speed
        padmode:AudioPadMode=AudioPadMode.Repeat, # How to pad if necessary
        constant:Number=0, # Value for `AudioPadMode.Constant`
        split:int|None=None # Apply transform to `split` items at a time. Use to prevent OOM.
    ):
        super().__init__(p=p)
        if padmode not in [vol for vol in AudioPadMode]:
            raise ValueError(f"`padmode` {padmode} is not valid")
        self.bernoulli = Bernoulli(.5)
        self.shift   = PitchShift(1., semitones, bins_per_octave, padmode, constant, split)
        self.stretch = TimeStretch(1., stretch_rates, padmode, constant, split)

    def encodes(self, x:TensorAudio) -> Tensor:
        self.shift_smpl = self.bernoulli.sample((x.shape[0],))
        self.stretchs_smpl = (self.shift_smpl - 1) *-1
        self.shift_idxs, self.stretch_idxs = self.shift_smpl.bool(), self.stretchs_smpl.bool()
        if self.shift_smpl.sum() > 0:
            x[self.shift_idxs] = self.shift(x[self.shift_idxs], split_idx=0)
        if self.stretchs_smpl.sum() > 0:
            x[self.stretch_idxs] = self.stretch(x[self.stretch_idxs], split_idx=0)
        return x

# Cell
class TimeMasking(BatchRandTransform):
    "Randomly apply masking to a `TensorSpec` or `TensorMelSpec` in the time domain"
    split_idx, order = 0, 81
    def __init__(self,
        p:float=0.25, # Per-item probability
        max_mask:float=0.2, # Maximum possible length of the mask  [0, max_mask)
        iid_masks:bool=True, # Apply different masks to each example/channel in the batch
        mask_value:int|None=0 # If None, random value between batch min and max
    ):
        store_attr(but='p')
        super().__init__(p=p)

    def encodes(self, x:TensorSpec|TensorMelSpec) -> Tensor:
        mask_param = int(self.max_mask * x.shape[-1])
        mv = random.randint(int(x.min), int(x.max)) if self.mask_value is None else self.mask_value
        if self.iid_masks and x.dim() == 4:
            return TAF.mask_along_axis_iid(x, mask_param, mv, 3)
        else:
            return TAF.mask_along_axis(x, mask_param, mv, 2)

# Cell
class FrequencyMasking(BatchRandTransform):
    "Randpmly apply masking to a `TensorSpec` or `TensorMelSpec` in the frequency domain"
    split_idx, order = 0, 82
    def __init__(self,
        p:float=0.25, # Per-item probability
        max_mask:float=0.2, # Maximum possible length of the mask [0, max_mask)
        iid_masks:bool=True, # Apply different masks to each example/channel in the batch
        mask_value:int|None=0 # If None, random value between batch min and max
    ):
        store_attr(but='p')
        super().__init__(p=p)

    def encodes(self, x:TensorSpec|TensorMelSpec) -> Tensor:
        mask_param = int(self.max_mask * x.shape[-2])
        mv = random.randint(int(x.min), int(x.max)) if self.mask_value is None else self.mask_value
        if self.iid_masks and x.dim() == 4:
            return TAF.mask_along_axis_iid(x, mask_param, mv, 2)
        else:
            return TAF.mask_along_axis(x, mask_param, mv, 1)

# Cell
class AmplitudeToDBMode(Enum):
    "All AmplitudeToDB modes as attributes to get tab-completion and typo-proofing",
    Power = 'power'
    Magnitude = 'magnitude'

# Cell
class AmplitudeToDB(DisplayedTransform):
    "Turn a `TensorSpec` or `TensorMelSpec` from the power/amplitude scale to the decibel scale"
    order = 98
    def __init__(self,
        top_db:float|None=None, # Minimum negative cut-off in decibels. A reasonable number is 80.
        mode:AmplitudeToDBMode=AmplitudeToDBMode.Power # Power is elementwise square of magnitude
    ):
        self.amdb = tatfms.AmplitudeToDB(mode.value, top_db)

    def encodes(self, x:TensorSpec|TensorMelSpec) -> Tensor:
        return self.amdb(x)

    def to(self, *args, **kwargs):
        device, dtype, non_blocking, convert_to_format = torch._C._nn._parse_to(*args, **kwargs)
        self.amdb.to(device)

# Cell
class AudioNormalize(DisplayedTransform):
    "Normalize/denorm batch of `TensorSpec` or `TensorMelSpec`."
    parameters,order = L('mean', 'std'),99
    def __init__(self, mean=None, std=None, n_spec=1, axes=(0,2,3)):
        store_attr()
        self.c = 0

    @classmethod
    def from_stats(cls, mean, std, n_spec=1, dim=1, ndim=4, cuda=True):
        return cls(*broadcast_vec(dim, ndim, mean, std, cuda=cuda), n_spec=n_spec)

    def setups(self, dl:DataLoader):
        if self.mean is None or self.std is None:
            x,*_ = dl.one_batch()
            self.mean,self.std = x.mean(self.axes, keepdim=True),x.std(self.axes, keepdim=True)+1e-7

    def encodes(self, x:TensorSpec|TensorMelSpec):
        if self.n_spec > 1:
            i = self.c*x.channels
            j = (self.c+1)*x.channels
            self.c = self.c + 1 if self.c < self.n_spec-1 else 0
            return (x-self.mean[:,i:j,...]) / self.std[:,i:j,...]
        else:
            return (x-self.mean) / self.std

    # ToDo: figure out how to handle decodes
    # def decodes(self, x:TensorSpec|TensorMelSpec):
    #     f = to_cpu if x.device.type=='cpu' else noop
    #     return (x*f(self.std) + f(self.mean))


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/vision.models.xresnet.ipynb (unless otherwise specified).

__all__ = ['ResBlock', 'ResNeXtBlock', 'SEBlock', 'SEResNeXtBlock', 'ECABlock', 'ECAResNeXtBlock', 'SABlock',
           'SAResNeXtBlock', 'TABlock', 'TAResNeXtBlock', 'XResNet', 'xresnet18', 'xresnet34', 'xresnet50',
           'xresnet101', 'xresnext18', 'xresnext34', 'xresnext50', 'xresnext101', 'xse_resnet18', 'xse_resnet34',
           'xse_resnet50', 'xse_resnet101', 'xse_resnext18', 'xse_resnext34', 'xse_resnext50', 'xse_resnext101',
           'xeca_resnet18', 'xeca_resnet34', 'xeca_resnet50', 'xeca_resnet101', 'xeca_resnext18', 'xeca_resnext34',
           'xeca_resnext50', 'xeca_resnext101', 'xsa_resnet18', 'xsa_resnet34', 'xsa_resnet50', 'xsa_resnet101',
           'xsa_resnext18', 'xsa_resnext34', 'xsa_resnext50', 'xsa_resnext101', 'xta_resnet18', 'xta_resnet34',
           'xta_resnet50', 'xta_resnet101', 'xta_resnext18', 'xta_resnext34', 'xta_resnext50', 'xta_resnext101']

# Cell
from torchvision.ops.stochastic_depth import StochasticDepth

from fastai.basics import defaults
from fastai.layers import ConvLayer, NormType, SimpleSelfAttention, AvgPool, SEModule, MaxPool, AdaptiveAvgPool, Flatten
from fastai.vision.models.xresnet import init_cnn

from .attention_modules import *
from ...imports import *

# Cell
class ResBlock(Module):
    "Resnet block from `ni` to `nh` with `stride`"
    @delegates(ConvLayer.__init__)
    def __init__(self, expansion, ni, nf, stride=1, groups=1, attn_mod=None, nh1=None, nh2=None,
                 dw=False, g2=1, sa=False, sym=False, norm_type=NormType.Batch, act_cls=defaults.activation,
                 ndim=2, ks=3, block_pool=AvgPool, pool_first=True, stoch_depth=0, **kwargs):
        norm2 = (NormType.BatchZero if norm_type==NormType.Batch else
                 NormType.InstanceZero if norm_type==NormType.Instance else norm_type)
        if nh2 is None: nh2 = nf
        if nh1 is None: nh1 = nh2
        nf,ni = nf*expansion,ni*expansion
        k0 = dict(norm_type=norm_type, act_cls=act_cls, ndim=ndim, **kwargs)
        k1 = dict(norm_type=norm2, act_cls=None, ndim=ndim, **kwargs)
        convpath  = [ConvLayer(ni,  nh2, ks, stride=stride, groups=ni if dw else groups, **k0),
                     ConvLayer(nh2,  nf, ks, groups=g2, **k1)
        ] if expansion == 1 else [
                     ConvLayer(ni,  nh1, 1, **k0),
                     ConvLayer(nh1, nh2, ks, stride=stride, groups=nh1 if dw else groups, **k0),
                     ConvLayer(nh2,  nf, 1, groups=g2, **k1)]
        if attn_mod: convpath.append(attn_mod(nf))
        if sa: convpath.append(SimpleSelfAttention(nf,ks=1,sym=sym))
        self.convpath = nn.Sequential(*convpath)
        idpath = []
        if ni!=nf: idpath.append(ConvLayer(ni, nf, 1, act_cls=None, ndim=ndim, **kwargs))
        if stride!=1:
            idpath.insert((1,0)[pool_first], block_pool(stride, ndim=ndim, ceil_mode=True))
        self.idpath = nn.Sequential(*idpath)
        self.act = act_cls(inplace=True) if act_cls is defaults.activation else act_cls()
        self.depth = nn.Identity() if stoch_depth==0 else StochasticDepth(stoch_depth, 'batch')

    def forward(self, x): return self.act(self.depth(self.convpath(x)) + self.idpath(x))

# Cell
@delegates(ResBlock)
def ResNeXtBlock(expansion, ni, nf, groups=32, stride=1, base_width=4, **kwargs):
    w = math.floor(nf * (base_width / 64)) * groups
    return ResBlock(expansion, ni, nf, stride=stride, groups=groups, nh2=w, **kwargs)

# Cell
@delegates(ResBlock)
def SEBlock(expansion, ni, nf, groups=1, se_reduction=16, stride=1, se_act_cls=defaults.activation, **kwargs):
    "A Squeeze and Excitation `XResNet` Block. Can set `se_act_cls` seperately."
    attn_mod = partial(SEModule, reduction=se_reduction, act_cls=se_act_cls)
    return ResBlock(expansion, ni, nf, stride=stride, groups=groups, attn_mod=attn_mod, nh1=nf*2, nh2=nf*expansion, **kwargs)

# Cell
@delegates(ResBlock)
def SEResNeXtBlock(expansion, ni, nf, groups=32, se_reduction=16, stride=1, base_width=4, se_act_cls=defaults.activation, **kwargs):
    "A Squeeze and Excitation `XResNeXtBlock`. Can set `se_act_cls` seperately."
    w = math.floor(nf * (base_width / 64)) * groups
    attn_mod = partial(SEModule, reduction=se_reduction, act_cls=se_act_cls)
    return ResBlock(expansion, ni, nf, stride=stride, groups=groups, attn_mod=attn_mod, nh2=w, **kwargs)

# Cell
@delegates(ResBlock)
def ECABlock(expansion, ni, nf, groups=1, eca_ks=None, stride=1, **kwargs):
    "An Efficient Channel Attention `XResNet` Block"
    attn_mod = partial(ECA, ks=eca_ks)
    return ResBlock(expansion, ni, nf, stride=stride, groups=groups, attn_mod=attn_mod, nh1=nf*2, nh2=nf*expansion, **kwargs)

# Cell
@delegates(ResBlock)
def ECAResNeXtBlock(expansion, ni, nf, groups=32, eca_ks=None, stride=1, base_width=4, **kwargs):
    "An Efficient Channel Attention `XResNeXtBlock`"
    w = math.floor(nf * (base_width / 64)) * groups
    attn_mod = partial(ECA, ks=eca_ks)
    return ResBlock(expansion, ni, nf, stride=stride, groups=groups, attn_mod=attn_mod, nh2=w, **kwargs)

# Cell
@delegates(ResBlock)
def SABlock(expansion, ni, nf, groups=1, sa_grps=64, stride=1, **kwargs):
    "A Shuffle Attention `XResNet` Block"
    attn_mod = partial(ShuffleAttention, groups=sa_grps)
    return ResBlock(expansion, ni, nf, stride=stride, groups=groups, attn_mod=attn_mod, nh1=nf*2, nh2=nf*expansion, **kwargs)

# Cell
@delegates(ResBlock)
def SAResNeXtBlock(expansion, ni, nf, groups=32, sa_grps=64, stride=1, base_width=4, **kwargs):
    "A Shuffle Attention `XResNeXtBlock`"
    w = math.floor(nf * (base_width / 64)) * groups
    attn_mod = partial(ShuffleAttention, groups=sa_grps)
    return ResBlock(expansion, ni, nf, stride=stride, groups=groups, attn_mod=attn_mod, nh2=w, **kwargs)

# Cell
@delegates(ResBlock)
def TABlock(expansion, ni, nf, groups=1, ta_ks=7, stride=1, **kwargs):
    "A Triplet Attention `XResNet` Block"
    attn_mod = partial(TripletAttention, ks=ta_ks)
    return ResBlock(expansion, ni, nf, stride=stride, groups=groups, attn_mod=attn_mod, nh1=nf*2, nh2=nf*expansion, **kwargs)

# Cell
@delegates(ResBlock)
def TAResNeXtBlock(expansion, ni, nf, groups=32, ta_ks=7, stride=1, base_width=4, **kwargs):
    "A Triplet Attention `XResNeXtBlock`"
    w = math.floor(nf * (base_width / 64)) * groups
    attn_mod = partial(TripletAttention, ks=ta_ks)
    return ResBlock(expansion, ni, nf, stride=stride, groups=groups, attn_mod=attn_mod, nh2=w, **kwargs)

# Cell
class XResNet(nn.Sequential):
    "A flexible version of fastai's XResNet"
    @delegates(ResBlock)
    def __init__(self, block, expansion, layers, p=0.0, c_in=3, n_out=1000, stem_szs=(32,32,64),
                 block_szs=[64,128,256,512], widen=1.0, sa=False, act_cls=defaults.activation, ndim=2,
                 ks=3, stride=2, stem_layer=ConvLayer, stem_pool=MaxPool, head_pool=AdaptiveAvgPool,
                 custom_head=None, **kwargs):
        store_attr('block,expansion,act_cls,ndim,ks')
        if ks % 2 == 0: raise Exception('kernel size has to be odd!')
        stem_szs = [c_in, *stem_szs]
        stem = [stem_layer(stem_szs[i], stem_szs[i+1], ks=ks, stride=stride if i==0 else 1,
                           act_cls=act_cls, ndim=ndim)
                for i in range(3)]

        assert len(layers) == len(block_szs), 'Length of `layers` must match `block_szs`'
        block_szs = [int(o*widen) for o in block_szs]
        block_szs = [stem_szs[-1]//expansion] + block_szs
        stem_pool = stem_pool(ks=ks, stride=stride, padding=ks//2, ndim=ndim)
        if not is_listy(stem_pool): stem_pool = [stem_pool]
        blocks    = self._make_blocks(layers, block_szs, sa, stride, **kwargs)

        if custom_head:
            head = custom_head(block_szs[-1]*expansion, n_out)
            if not is_listy(head): head = [head]
            body = nn.Sequential(*stem, *stem_pool, *blocks)
            init_cnn(body)
            super().__init__(*list(body), *head)
        else:
            head = self._make_head(block_szs[-1]*expansion, head_pool, ndim, p, n_out)
            super().__init__(*stem, *stem_pool, *blocks, *head)
            init_cnn(self)

    def _make_blocks(self, layers, block_szs, sa, stride, **kwargs):
        return [self._make_layer(ni=block_szs[i], nf=block_szs[i+1], blocks=l,
                                 stride=1 if i==0 else stride, sa=sa and i==len(layers)-4, **kwargs)
                for i,l in enumerate(layers)]

    def _make_layer(self, ni, nf, blocks, stride, sa, **kwargs):
        return nn.Sequential(
            *[self.block(self.expansion, ni if i==0 else nf, nf, stride=stride if i==0 else 1,
                      sa=sa and i==(blocks-1), act_cls=self.act_cls, ndim=self.ndim, ks=self.ks, **kwargs)
              for i in range(blocks)])

    def _make_head(self, ni, head_pool, ndim, p, n_out):
        return [head_pool(sz=1, ndim=ndim), Flatten(), nn.Dropout(p), nn.Linear(ni, n_out)]

# Cell
@delegates(XResNet)
def xresnet18(n_out=1000, c_in=3, p=0.0, act_cls=defaults.activation, **kwargs):
    return XResNet(ResBlock, 1, [2, 2, 2, 2], n_out=n_out, **kwargs)

@delegates(XResNet)
def xresnet34(n_out=1000, **kwargs):
    return XResNet(ResBlock, 1, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xresnet50(n_out=1000, **kwargs):
    return XResNet(ResBlock, 4, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xresnet101(n_out=1000, **kwargs):
    return XResNet(ResBlock, 4, [3, 4, 23, 3], n_out=n_out, **kwargs)

# Cell
@delegates(XResNet)
def xresnext18(n_out=1000, **kwargs):
    return XResNet(ResNeXtBlock, 1, [2, 2, 2, 2], n_out=n_out, **kwargs)

@delegates(XResNet)
def xresnext34(n_out=1000, **kwargs):
    return XResNet(ResNeXtBlock, 1, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xresnext50(n_out=1000, **kwargs):
    return XResNet(ResNeXtBlock, 4, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xresnext101(n_out=1000, **kwargs):
    return XResNet(ResNeXtBlock, 4, [3, 4, 23, 3], n_out=n_out, **kwargs)

# Cell
@delegates(XResNet)
def xse_resnet18(n_out=1000, **kwargs):
    return XResNet(SEBlock, 1, [2, 2, 2, 2], n_out=n_out, **kwargs)

@delegates(XResNet)
def xse_resnet34(n_out=1000, **kwargs):
    return XResNet(SEBlock, 1, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xse_resnet50(n_out=1000, **kwargs):
    return XResNet(SEBlock, 4, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xse_resnet101(n_out=1000, **kwargs):
    return XResNet(SEBlock, 4, [3, 4, 23, 3], n_out=n_out, **kwargs)

# Cell
@delegates(XResNet)
def xse_resnext18(n_out=1000, **kwargs):
    return XResNet(SEResNeXtBlock, 1, [2, 2, 2, 2], n_out=n_out, **kwargs)

@delegates(XResNet)
def xse_resnext34(n_out=1000, **kwargs):
    return XResNet(SEResNeXtBlock, 1, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xse_resnext50(n_out=1000, **kwargs):
    return XResNet(SEResNeXtBlock, 4, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xse_resnext101(n_out=1000, **kwargs):
    return XResNet(SEResNeXtBlock, 4, [3, 4, 23, 3], n_out=n_out, **kwargs)

# Cell
@delegates(XResNet)
def xeca_resnet18(n_out=1000, **kwargs):
    return XResNet(ECABlock, 1, [2, 2, 2, 2], n_out=n_out, **kwargs)

@delegates(XResNet)
def xeca_resnet34(n_out=1000, **kwargs):
    return XResNet(ECABlock, 1, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xeca_resnet50(n_out=1000, **kwargs):
    return XResNet(ECABlock, 4, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xeca_resnet101(n_out=1000, **kwargs):
    return XResNet(ECABlock, 4, [3, 4, 23, 3], n_out=n_out, **kwargs)

# Cell
@delegates(XResNet)
def xeca_resnext18(n_out=1000, **kwargs):
    return XResNet(ECAResNeXtBlock, 1, [2, 2, 2, 2], n_out=n_out, **kwargs)

@delegates(XResNet)
def xeca_resnext34(n_out=1000, **kwargs):
    return XResNet(ECAResNeXtBlock, 1, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xeca_resnext50(n_out=1000, **kwargs):
    return XResNet(ECAResNeXtBlock, 4, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xeca_resnext101(n_out=1000, **kwargs):
    return XResNet(ECAResNeXtBlock, 4, [3, 4, 23, 3], n_out=n_out, **kwargs)

# Cell
@delegates(XResNet)
def xsa_resnet18(n_out=1000, **kwargs):
    return XResNet(SABlock, 1, [2, 2, 2, 2], n_out=n_out, sa_grps=32, **kwargs)

@delegates(XResNet)
def xsa_resnet34(n_out=1000, **kwargs):
    return XResNet(SABlock, 1, [3, 4, 6, 3], n_out=n_out, sa_grps=32, **kwargs)

@delegates(XResNet)
def xsa_resnet50(n_out=1000, **kwargs):
    return XResNet(SABlock, 4, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xsa_resnet101(n_out=1000, **kwargs):
    return XResNet(SABlock, 4, [3, 4, 23, 3], n_out=n_out, **kwargs)

# Cell
@delegates(XResNet)
def xsa_resnext18(n_out=1000, **kwargs):
    return XResNet(SAResNeXtBlock, 1, [2, 2, 2, 2], n_out=n_out, sa_grps=32, **kwargs)

@delegates(XResNet)
def xsa_resnext34(n_out=1000, **kwargs):
    return XResNet(SAResNeXtBlock, 1, [3, 4, 6, 3], n_out=n_out, sa_grps=32, **kwargs)

@delegates(XResNet)
def xsa_resnext50(n_out=1000, **kwargs):
    return XResNet(SAResNeXtBlock, 4, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xsa_resnext101(n_out=1000, **kwargs):
    return XResNet(SAResNeXtBlock, 4, [3, 4, 23, 3], n_out=n_out, **kwargs)

# Cell
@delegates(XResNet)
def xta_resnet18(n_out=1000, **kwargs):
    return XResNet(TABlock, 1, [2, 2, 2, 2], n_out=n_out, **kwargs)

@delegates(XResNet)
def xta_resnet34(n_out=1000, **kwargs):
    return XResNet(TABlock, 1, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xta_resnet50(n_out=1000, **kwargs):
    return XResNet(TABlock, 4, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xta_resnet101(n_out=1000, **kwargs):
    return XResNet(TABlock, 4, [3, 4, 23, 3], n_out=n_out, **kwargs)

# Cell
@delegates(XResNet)
def xta_resnext18(n_out=1000, **kwargs):
    return XResNet(TAResNeXtBlock, 1, [2, 2, 2, 2], n_out=n_out, sa_grps=32, **kwargs)

@delegates(XResNet)
def xta_resnext34(n_out=1000, **kwargs):
    return XResNet(TAResNeXtBlock, 1, [3, 4, 6, 3], n_out=n_out, sa_grps=32, **kwargs)

@delegates(XResNet)
def xta_resnext50(n_out=1000, **kwargs):
    return XResNet(TAResNeXtBlock, 4, [3, 4, 6, 3], n_out=n_out, **kwargs)

@delegates(XResNet)
def xta_resnext101(n_out=1000, **kwargs):
    return XResNet(TAResNeXtBlock, 4, [3, 4, 23, 3], n_out=n_out, **kwargs)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/callback.tracker.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['TerminateOnTrainNaN', 'SaveModelAtEnd']

# Cell
#nbdev_comment from __future__ import annotations

from fastai.callback.core import Callback, CancelFitException
from fastai.callback.tracker import SaveModelCallback, TrackerCallback

from ..imports import *

# Cell
class TerminateOnTrainNaN(Callback):
    "A `Callback` that terminates training if the training loss is NaN and ignores valid loss."
    order,run_valid=-9,False
    def after_batch(self):
        "Test if `last_loss` is NaN and interrupts training."
        if torch.isinf(self.loss) or torch.isnan(self.loss): raise CancelFitException

# Cell
class SaveModelAtEnd(SaveModelCallback):
    "A `SaveModelCallback` which only saves the model at the end so loggers can find it."
    order = TrackerCallback.order+1
    def __init__(self,
        fname='model', # Model filename
        with_opt=False # Include optimizer state
    ):
        # keep track of file path for loggers
        self.last_saved_path = None
        store_attr()

    def before_fit(self):
        pass

    def after_epoch(self):
        pass

    def after_fit(self, **kwargs):
        self.last_saved_path = self.learn.save(f'{self.fname}', with_opt=self.with_opt)